# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2021

from ..log import logger
import re


class Traceparent:
    SPECIFICATION_VERSION = "00"
    TRACEPARENT_REGEX = re.compile("^[0-9a-f]{2}-(?!0{32})([0-9a-f]{32})-(?!0{16})([0-9a-f]{16})-[0-9a-f]{2}")

    def validate(self, traceparent):
        """
        Method used to validate the traceparent header
        :param traceparent: string
        :return: traceparent or None
        """
        try:
            if self.TRACEPARENT_REGEX.match(traceparent):
                return traceparent
        except Exception:
            logger.debug("traceparent does not follow version {} specification".format(self.SPECIFICATION_VERSION))
        return None

    @staticmethod
    def get_traceparent_fields(traceparent):
        """
        Parses the validated traceparent header into its fields and returns the fields
        :param traceparent: the original validated traceparent header
        :return: version, trace_id, parent_id, trace_flags
        """
        try:
            traceparent_properties = traceparent.split("-")
            version = traceparent_properties[0]
            trace_id = traceparent_properties[1]
            parent_id = traceparent_properties[2]
            trace_flags = traceparent_properties[3]
            return version, trace_id, parent_id, trace_flags
        except Exception:  # This method is intended to be called with a version 00 validated traceparent
            # This exception handling is added just for making sure we do not throw any unhandled exception
            # if somebody calls the method in the future without a validated traceparent
            return None, None, None, None

    def update_traceparent(self, traceparent, in_trace_id, in_span_id, level):
        """
        This method updates the traceparent header or generates one if there was no traceparent incoming header or it
        was invalid
        :param traceparent: the original validated traceparent header
        :param in_trace_id: instana trace id, used when there is no preexisting trace_id from the traceparent header
        :param in_span_id: instana span id, used to update the parent id of the traceparent header
        :param level: instana level, used to determine the value of sampled flag of the traceparent header
        :return: the updated traceparent header
        """
        mask = 1 << 0
        trace_flags = 0
        if traceparent is None:  # modify the trace_id part only when it was not present at all
            trace_id = in_trace_id.zfill(32)
            version = self.SPECIFICATION_VERSION
        else:
            version, trace_id, _, trace_flags = self.get_traceparent_fields(traceparent)
            trace_flags = int(trace_flags, 16)

        parent_id = in_span_id.zfill(16)
        trace_flags = (trace_flags & ~mask) | ((level << 0) & mask)
        trace_flags = format(trace_flags, '0>2x')

        traceparent = "{version}-{traceid}-{parentid}-{trace_flags}".format(version=version,
                                                                        traceid=trace_id,
                                                                        parentid=parent_id,
                                                                        trace_flags=trace_flags)
        return traceparent



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

"""
Host Collector: Manages the periodic collection of metrics & snapshot data
"""
from time import time
from ..log import logger
from .base import BaseCollector
from ..util import DictionaryOfStan
from .helpers.runtime import RuntimeHelper


class HostCollector(BaseCollector):
    """ Collector for AWS Fargate """
    def __init__(self, agent):
        super(HostCollector, self).__init__(agent)
        logger.debug("Loading Host Collector")

        # Indicates if this Collector has all requirements to run successfully
        self.ready_to_start = True

        # Populate the collection helpers
        self.helpers.append(RuntimeHelper(self))

    def start(self):
        if self.ready_to_start is False:
            logger.warning("Host Collector is missing requirements and cannot monitor this environment.")
            return

        super(HostCollector, self).start()

    def prepare_and_report_data(self):
        """
        We override this method from the base class so that we can handle the wait4init
        state machine case.
        """
        try:
            if self.agent.machine.fsm.current == "wait4init":
                # Test the host agent if we're ready to send data
                if self.agent.is_agent_ready():
                    if self.agent.machine.fsm.current != "good2go":
                        logger.debug("Agent is ready.  Getting to work.")
                        self.agent.machine.fsm.ready()
                else:
                    return

            if self.agent.machine.fsm.current == "good2go" and self.agent.is_timed_out():
                logger.info("The Instana host agent has gone offline or is no longer reachable for > 1 min.  Will retry periodically.")
                self.agent.reset()
        except Exception:
            logger.debug('Harmless state machine thread disagreement.  Will self-correct on next timer cycle.')

        super(HostCollector, self).prepare_and_report_data()

    def should_send_snapshot_data(self):
        delta = int(time()) - self.snapshot_data_last_sent
        if delta > self.snapshot_data_interval:
            return True
        return False

    def prepare_payload(self):
        payload = DictionaryOfStan()
        payload["spans"] = []
        payload["profiles"] = []
        payload["metrics"]["plugins"] = []

        try:
            if not self.span_queue.empty():
                payload["spans"] = self.queued_spans()

            if not self.profile_queue.empty():
                payload["profiles"] = self.queued_profiles()

            with_snapshot = self.should_send_snapshot_data()

            plugins = []
            for helper in self.helpers:
                plugins.extend(helper.collect_metrics(with_snapshot=with_snapshot))

            payload["metrics"]["plugins"] = plugins

            if with_snapshot is True:
                self.snapshot_data_last_sent = int(time())
        except Exception:
            logger.debug("non-fatal prepare_payload:", exc_info=True)

        return payload



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

from __future__ import absolute_import

from instana.log import logger
from instana.propagators.base_propagator import BasePropagator


class HTTPPropagator(BasePropagator):
    """
    Instana Propagator for Format.HTTP_HEADERS.

    The HTTP_HEADERS format deals with key-values with string to string mapping.
    The character set should be restricted to HTTP compatible.
    """

    def __init__(self):
        super(HTTPPropagator, self).__init__()

    def inject(self, span_context, carrier, disable_w3c_trace_context=False):
        try:
            trace_id = span_context.trace_id
            span_id = span_context.span_id
            level = span_context.level

            if disable_w3c_trace_context:
                traceparent, tracestate = [None] * 2
            else:
                traceparent, tracestate = self._get_participating_trace_context(span_context)

            if isinstance(carrier, dict) or hasattr(carrier, "__dict__"):
                if traceparent and tracestate:
                    carrier[self.HEADER_KEY_TRACEPARENT] = traceparent
                    carrier[self.HEADER_KEY_TRACESTATE] = tracestate
                carrier[self.HEADER_KEY_T] = trace_id
                carrier[self.HEADER_KEY_S] = span_id
                carrier[self.HEADER_KEY_L] = "1"
            elif isinstance(carrier, list):
                if traceparent and tracestate:
                    carrier.append((self.HEADER_KEY_TRACEPARENT, traceparent))
                    carrier.append((self.HEADER_KEY_TRACESTATE, tracestate))
                carrier.append((self.HEADER_KEY_T, trace_id))
                carrier.append((self.HEADER_KEY_S, span_id))
                carrier.append((self.HEADER_KEY_L, "1"))
            elif hasattr(carrier, '__setitem__'):
                if traceparent and tracestate:
                    carrier.__setitem__(self.HEADER_KEY_TRACEPARENT, traceparent)
                    carrier.__setitem__(self.HEADER_KEY_TRACESTATE, tracestate)
                carrier.__setitem__(self.HEADER_KEY_T, trace_id)
                carrier.__setitem__(self.HEADER_KEY_S, span_id)
                carrier.__setitem__(self.HEADER_KEY_L, "1")
            else:
                raise Exception("Unsupported carrier type", type(carrier))

        except Exception:
            logger.debug("inject error:", exc_info=True)













# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2021


import uvicorn
from ...helpers import testenv
from instana.log import logger

testenv["sanic_port"] = 1337
testenv["sanic_server"] = ("http://127.0.0.1:" + str(testenv["sanic_port"]))


def launch_sanic():
    from .server import app
    from instana.singletons import agent

    # Hack together a manual custom headers list; We'll use this in tests
    agent.options.extra_http_headers = [u'X-Capture-This', u'X-Capture-That']

    uvicorn.run(app, host='127.0.0.1', port=testenv['sanic_port'], log_level="critical")



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

"""
The Instana agent (for AWS Lambda functions) that manages
monitoring state and reporting that data.
"""
import time
from ..log import logger
from ..util import to_json
from .base import BaseAgent
from ..version import VERSION
from ..collector.aws_lambda import AWSLambdaCollector
from ..options import AWSLambdaOptions


class AWSLambdaAgent(BaseAgent):
    """ In-process agent for AWS Lambda """
    def __init__(self):
        super(AWSLambdaAgent, self).__init__()

        self.collector = None
        self.options = AWSLambdaOptions()
        self.report_headers = None
        self._can_send = False

        # Update log level from what Options detected
        self.update_log_level()

        logger.info("Stan is on the AWS Lambda scene.  Starting Instana instrumentation version: %s", VERSION)

        if self._validate_options():
            self._can_send = True
            self.collector = AWSLambdaCollector(self)
            self.collector.start()
        else:
            logger.warning("Required INSTANA_AGENT_KEY and/or INSTANA_ENDPOINT_URL environment variables not set.  "
                           "We will not be able monitor this function.")

    def can_send(self):
        """
        Are we in a state where we can send data?
        @return: Boolean
        """
        return self._can_send

    def get_from_structure(self):
        """
        Retrieves the From data that is reported alongside monitoring data.
        @return: dict()
        """
        return {'hl': True, 'cp': 'aws', 'e': self.collector.get_fq_arn()}

    def report_data_payload(self, payload):
        """
        Used to report metrics and span data to the endpoint URL in self.options.endpoint_url
        """
        response = None
        try:
            if self.report_headers is None:
                # Prepare request headers
                self.report_headers = dict()
                self.report_headers["Content-Type"] = "application/json"
                self.report_headers["X-Instana-Host"] = self.collector.get_fq_arn()
                self.report_headers["X-Instana-Key"] = self.options.agent_key

            self.report_headers["X-Instana-Time"] = str(round(time.time() * 1000))

            response = self.client.post(self.__data_bundle_url(),
                                        data=to_json(payload),
                                        headers=self.report_headers,
                                        timeout=self.options.timeout,
                                        verify=self.options.ssl_verify,
                                        proxies=self.options.endpoint_proxy)

            if 200 <= response.status_code < 300:
                logger.debug("report_data_payload: Instana responded with status code %s", response.status_code)
            else:
                logger.info("report_data_payload: Instana responded with status code %s", response.status_code)
        except Exception as exc:
            logger.debug("report_data_payload: connection error (%s)", type(exc))

        return response

    def _validate_options(self):
        """
        Validate that the options used by this Agent are valid.  e.g. can we report data?
        """
        return self.options.endpoint_url is not None and self.options.agent_key is not None

    def __data_bundle_url(self):
        """
        URL for posting metrics to the host agent.  Only valid when announced.
        """
        return "%s/bundle" % self.options.endpoint_url



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import re

def sql_sanitizer(sql):
    """
    Removes values from valid SQL statements and returns a stripped version.

    :param sql: The SQL statement to be sanitized
    :return: String - A sanitized SQL statement without values.
    """
    return regexp_sql_values.sub('?', sql)


# Used by sql_sanitizer
regexp_sql_values = re.compile(r"('[\s\S][^']*'|\d*\.\d+|\d+|NULL)")


# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

"""
Module to handle the work related to the many AWS Lambda Triggers.
"""
import gzip
import json
import base64
from io import BytesIO
import opentracing as ot

from ...log import logger

STR_LAMBDA_TRIGGER = 'lambda.trigger'


def get_context(tracer, event):
    # TODO: Search for more types of trigger context
    is_proxy_event = is_api_gateway_proxy_trigger(event) or \
        is_api_gateway_v2_proxy_trigger(event) or \
        is_application_load_balancer_trigger(event)

    if is_proxy_event:
        return tracer.extract(ot.Format.HTTP_HEADERS, event.get('headers', {}), disable_w3c_trace_context=True)

    return tracer.extract(ot.Format.HTTP_HEADERS, event, disable_w3c_trace_context=True)


def is_api_gateway_proxy_trigger(event):
    for key in ["resource", "path", "httpMethod"]:
        if key not in event:
            return False
    return True


def is_api_gateway_v2_proxy_trigger(event):
    for key in ["version", "requestContext"]:
        if key not in event:
            return False

    if event["version"] != "2.0":
        return False

    for key in ["apiId", "stage", "http"]:
        if key not in event["requestContext"]:
            return False

    return True

def is_application_load_balancer_trigger(event):
    if 'requestContext' in event and 'elb' in event['requestContext']:
        return True
    return False


def is_cloudwatch_trigger(event):
    if "source" in event and 'detail-type' in event:
        if event["source"] == 'aws.events' and event['detail-type'] == 'Scheduled Event':
            return True
    return False


def is_cloudwatch_logs_trigger(event):
    if hasattr(event, 'get') and event.get("awslogs", False) is not False:
        return True
    else:
        return False


def is_s3_trigger(event):
    if "Records" in event:
        if len(event["Records"]) > 0 and event["Records"][0]["eventSource"] == 'aws:s3':
            return True
    return False


def is_sqs_trigger(event):
    if "Records" in event:
        if len(event["Records"]) > 0 and event["Records"][0]["eventSource"] == 'aws:sqs':
            return True
    return False


def read_http_query_params(event):
    """
    Used to parse the Lambda QueryString formats.

    @param event: lambda event dict
    @return: String in the form of "a=b&c=d"
    """
    params = []
    try:
        if event is None or type(event) is not dict:
            return ""

        mvqsp = event.get('multiValueQueryStringParameters', None)
        qsp = event.get('queryStringParameters', None)

        if mvqsp is not None and type(mvqsp) is dict:
            for key in mvqsp:
                params.append("%s=%s" % (key, mvqsp[key]))
            return "&".join(params)
        elif qsp is not None and type(qsp) is dict:
            for key in qsp:
                params.append("%s=%s" % (key, qsp[key]))
            return "&".join(params)
        else:
            return ""
    except Exception:
        logger.debug("read_http_query_params: ", exc_info=True)
        return ""


def capture_extra_headers(event, span, extra_headers):
    """
    Capture the headers specified in `extra_headers` from `event` and log them
    as a tag in the span.

    @param event: the lambda event
    @param span: the lambda entry span
    @param extra_headers: a list of http headers to capture
    @return: None
    """
    try:
        event_headers = event.get("headers", None)

        if event_headers is not None:
            for custom_header in extra_headers:
                for key in event_headers:
                    if key.lower() == custom_header.lower():
                        span.set_tag("http.header.%s" % custom_header, event_headers[key])
    except Exception:
        logger.debug("capture_extra_headers: ", exc_info=True)


def enrich_lambda_span(agent, span, event, context):
    """
    Extract the required information about this Lambda run (and the trigger) and store the data
    on `span`.

    @param agent: the AWSLambdaAgent in use
    @param span: the Lambda entry span
    @param event: the lambda handler event
    @param context: the lambda handler context
    @return: None
    """
    try:
        span.set_tag('lambda.arn', agent.collector.get_fq_arn())
        span.set_tag('lambda.name', context.function_name)
        span.set_tag('lambda.version', context.function_version)

        if event is None or type(event) is not dict:
            logger.debug("enrich_lambda_span: bad event %s", type(event))
            return

        if is_api_gateway_proxy_trigger(event):
            logger.debug("Detected as API Gateway Proxy Trigger")
            span.set_tag(STR_LAMBDA_TRIGGER, 'aws:api.gateway')
            span.set_tag('http.method', event["httpMethod"])
            span.set_tag('http.url', event["path"])
            span.set_tag('http.path_tpl', event["resource"])
            span.set_tag('http.params', read_http_query_params(event))

            if agent.options.extra_http_headers is not None:
                capture_extra_headers(event, span, agent.options.extra_http_headers)

        elif is_api_gateway_v2_proxy_trigger(event):
            logger.debug("Detected as API Gateway v2.0 Proxy Trigger")

            reqCtx = event["requestContext"]

            # trim optional HTTP method prefix
            route_path = event["routeKey"].split(" ", 2)[-1]

            span.set_tag(STR_LAMBDA_TRIGGER, 'aws:api.gateway')
            span.set_tag('http.method', reqCtx["http"]["method"])
            span.set_tag('http.url', reqCtx["http"]["path"])
            span.set_tag('http.path_tpl', route_path)
            span.set_tag('http.params', read_http_query_params(event))

            if agent.options.extra_http_headers is not None:
                capture_extra_headers(event, span, agent.options.extra_http_headers)

        elif is_application_load_balancer_trigger(event):
            logger.debug("Detected as Application Load Balancer Trigger")
            span.set_tag(STR_LAMBDA_TRIGGER, 'aws:application.load.balancer')
            span.set_tag('http.method', event["httpMethod"])
            span.set_tag('http.url', event["path"])
            span.set_tag('http.params', read_http_query_params(event))

            if agent.options.extra_http_headers is not None:
                capture_extra_headers(event, span, agent.options.extra_http_headers)

        elif is_cloudwatch_trigger(event):
            logger.debug("Detected as Cloudwatch Trigger")
            span.set_tag(STR_LAMBDA_TRIGGER, 'aws:cloudwatch.events')
            span.set_tag('data.lambda.cw.events.id', event['id'])

            resources = event['resources']
            resource_count = len(event['resources'])
            if resource_count > 3:
                resources = event['resources'][:3]
                span.set_tag('lambda.cw.events.more', True)
            else:
                span.set_tag('lambda.cw.events.more', False)

            report = []
            for item in resources:
                if len(item) > 200:
                    item = item[:200]
                report.append(item)
            span.set_tag('lambda.cw.events.resources', report)

        elif is_cloudwatch_logs_trigger(event):
            logger.debug("Detected as Cloudwatch Logs Trigger")
            span.set_tag(STR_LAMBDA_TRIGGER, 'aws:cloudwatch.logs')

            try:
                if 'awslogs' in event and 'data' in event['awslogs']:
                    data = event['awslogs']['data']
                    decoded_data = base64.b64decode(data)
                    decompressed_data = gzip.GzipFile(fileobj=BytesIO(decoded_data)).read()
                    log_data = json.loads(decompressed_data.decode('utf-8'))

                    span.set_tag('lambda.cw.logs.group', log_data.get('logGroup', None))
                    span.set_tag('lambda.cw.logs.stream', log_data.get('logStream', None))
                    if len(log_data['logEvents']) > 3:
                        span.set_tag('lambda.cw.logs.more', True)
                        events = log_data['logEvents'][:3]
                    else:
                        events = log_data['logEvents']

                    event_data = []
                    for item in events:
                        msg = item.get('message', None)
                        if len(msg) > 200:
                            msg = msg[:200]
                        event_data.append(msg)
                    span.set_tag('lambda.cw.logs.events', event_data)
            except Exception as e:
                span.set_tag('lambda.cw.logs.decodingError', repr(e))
        elif is_s3_trigger(event):
            logger.debug("Detected as S3 Trigger")
            span.set_tag(STR_LAMBDA_TRIGGER, 'aws:s3')

            if "Records" in event:
                events = []
                for item in event["Records"][:3]:
                    bucket_name = "Unknown"
                    if "s3" in item and "bucket" in item["s3"]:
                        bucket_name = item["s3"]["bucket"]["name"]

                    object_name = ""
                    if "s3" in item and "object" in item["s3"]:
                        object_name = item["s3"]["object"].get("key", "Unknown")

                    if len(object_name) > 200:
                        object_name = object_name[:200]

                    events.append({"event": item['eventName'],
                                   "bucket": bucket_name,
                                   "object": object_name})
                span.set_tag('lambda.s3.events', events)

        elif is_sqs_trigger(event):
            logger.debug("Detected as SQS Trigger")
            span.set_tag(STR_LAMBDA_TRIGGER, 'aws:sqs')

            if "Records" in event:
                events = []
                for item in event["Records"][:3]:
                    events.append({'queue': item['eventSourceARN']})
                span.set_tag('lambda.sqs.messages', events)
        else:
            logger.debug("Detected as Unknown Trigger: %s" % event)
            span.set_tag(STR_LAMBDA_TRIGGER, 'unknown')

    except Exception:
        logger.debug("enrich_lambda_span: ", exc_info=True)



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import opentracing


def test_tracer_basics():
    assert hasattr(opentracing.tracer, "start_span")
    assert hasattr(opentracing.tracer, "inject")
    assert hasattr(opentracing.tracer, "extract")



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

from __future__ import absolute_import

import os
import json
import unittest

from instana.tracer import InstanaTracer
from instana.recorder import StanRecorder
from instana.agent.host import HostAgent
from instana.singletons import get_agent, set_agent, get_tracer, set_tracer


class TestHostCollector(unittest.TestCase):
    def __init__(self, methodName='runTest'):
        super(TestHostCollector, self).__init__(methodName)
        self.agent = None
        self.span_recorder = None
        self.tracer = None

        self.original_agent = get_agent()
        self.original_tracer = get_tracer()

    def setUp(self):
        pass

    def tearDown(self):
        """ Reset all environment variables of consequence """
        if "AWS_EXECUTION_ENV" in os.environ:
            os.environ.pop("AWS_EXECUTION_ENV")
        if "INSTANA_EXTRA_HTTP_HEADERS" in os.environ:
            os.environ.pop("INSTANA_EXTRA_HTTP_HEADERS")
        if "INSTANA_ENDPOINT_URL" in os.environ:
            os.environ.pop("INSTANA_ENDPOINT_URL")
        if "INSTANA_AGENT_KEY" in os.environ:
            os.environ.pop("INSTANA_AGENT_KEY")
        if "INSTANA_ZONE" in os.environ:
            os.environ.pop("INSTANA_ZONE")
        if "INSTANA_TAGS" in os.environ:
            os.environ.pop("INSTANA_TAGS")
        if "INSTANA_DISABLE_METRICS_COLLECTION" in os.environ:
            os.environ.pop("INSTANA_DISABLE_METRICS_COLLECTION")

        set_agent(self.original_agent)
        set_tracer(self.original_tracer)

    def create_agent_and_setup_tracer(self):
        self.agent = HostAgent()
        self.span_recorder = StanRecorder(self.agent)
        self.tracer = InstanaTracer(recorder=self.span_recorder)
        set_agent(self.agent)
        set_tracer(self.tracer)

    def test_prepare_payload_basics(self):
        self.create_agent_and_setup_tracer()

        payload = self.agent.collector.prepare_payload()
        assert (payload)

        assert (len(payload.keys()) == 3)
        assert ('spans' in payload)
        assert (isinstance(payload['spans'], list))
        assert (len(payload['spans']) == 0)
        assert ('metrics' in payload)
        assert (len(payload['metrics'].keys()) == 1)
        assert ('plugins' in payload['metrics'])
        assert (isinstance(payload['metrics']['plugins'], list))
        assert (len(payload['metrics']['plugins']) == 1)

        python_plugin = payload['metrics']['plugins'][0]
        assert python_plugin['name'] == 'com.instana.plugin.python'
        assert python_plugin['entityId'] == str(os.getpid())
        assert 'data' in python_plugin
        assert 'snapshot' in python_plugin['data']
        assert 'metrics' in python_plugin['data']

        # Validate that all metrics are reported on the first run
        assert 'ru_utime' in python_plugin['data']['metrics']
        assert type(python_plugin['data']['metrics']['ru_utime']) in [float, int]
        assert 'ru_stime' in python_plugin['data']['metrics']
        assert type(python_plugin['data']['metrics']['ru_stime']) in [float, int]
        assert 'ru_maxrss' in python_plugin['data']['metrics']
        assert type(python_plugin['data']['metrics']['ru_maxrss']) in [float, int]
        assert 'ru_ixrss' in python_plugin['data']['metrics']
        assert type(python_plugin['data']['metrics']['ru_ixrss']) in [float, int]
        assert 'ru_idrss' in python_plugin['data']['metrics']
        assert type(python_plugin['data']['metrics']['ru_idrss']) in [float, int]
        assert 'ru_isrss' in python_plugin['data']['metrics']
        assert type(python_plugin['data']['metrics']['ru_isrss']) in [float, int]
        assert 'ru_minflt' in python_plugin['data']['metrics']
        assert type(python_plugin['data']['metrics']['ru_minflt']) in [float, int]
        assert 'ru_majflt' in python_plugin['data']['metrics']
        assert type(python_plugin['data']['metrics']['ru_majflt']) in [float, int]
        assert 'ru_nswap' in python_plugin['data']['metrics']
        assert type(python_plugin['data']['metrics']['ru_nswap']) in [float, int]
        assert 'ru_inblock' in python_plugin['data']['metrics']
        assert type(python_plugin['data']['metrics']['ru_inblock']) in [float, int]
        assert 'ru_oublock' in python_plugin['data']['metrics']
        assert type(python_plugin['data']['metrics']['ru_oublock']) in [float, int]
        assert 'ru_msgsnd' in python_plugin['data']['metrics']
        assert type(python_plugin['data']['metrics']['ru_msgsnd']) in [float, int]
        assert 'ru_msgrcv' in python_plugin['data']['metrics']
        assert type(python_plugin['data']['metrics']['ru_msgrcv']) in [float, int]
        assert 'ru_nsignals' in python_plugin['data']['metrics']
        assert type(python_plugin['data']['metrics']['ru_nsignals']) in [float, int]
        assert 'ru_nvcsw' in python_plugin['data']['metrics']
        assert type(python_plugin['data']['metrics']['ru_nvcsw']) in [float, int]
        assert 'ru_nivcsw' in python_plugin['data']['metrics']
        assert type(python_plugin['data']['metrics']['ru_nivcsw']) in [float, int]
        assert 'alive_threads' in python_plugin['data']['metrics']
        assert type(python_plugin['data']['metrics']['alive_threads']) in [float, int]
        assert 'dummy_threads' in python_plugin['data']['metrics']
        assert type(python_plugin['data']['metrics']['dummy_threads']) in [float, int]
        assert 'daemon_threads' in python_plugin['data']['metrics']
        assert type(python_plugin['data']['metrics']['daemon_threads']) in [float, int]

        assert 'gc' in python_plugin['data']['metrics']
        assert isinstance(python_plugin['data']['metrics']['gc'], dict)
        assert 'collect0' in python_plugin['data']['metrics']['gc']
        assert type(python_plugin['data']['metrics']['gc']['collect0']) in [float, int]
        assert 'collect1' in python_plugin['data']['metrics']['gc']
        assert type(python_plugin['data']['metrics']['gc']['collect1']) in [float, int]
        assert 'collect2' in python_plugin['data']['metrics']['gc']
        assert type(python_plugin['data']['metrics']['gc']['collect2']) in [float, int]
        assert 'threshold0' in python_plugin['data']['metrics']['gc']
        assert type(python_plugin['data']['metrics']['gc']['threshold0']) in [float, int]
        assert 'threshold1' in python_plugin['data']['metrics']['gc']
        assert type(python_plugin['data']['metrics']['gc']['threshold1']) in [float, int]
        assert 'threshold2' in python_plugin['data']['metrics']['gc']
        assert type(python_plugin['data']['metrics']['gc']['threshold2']) in [float, int]

    def test_prepare_payload_basics_disable_runtime_metrics(self):
        os.environ["INSTANA_DISABLE_METRICS_COLLECTION"] = "TRUE"
        self.create_agent_and_setup_tracer()

        payload = self.agent.collector.prepare_payload()
        assert (payload)

        assert (len(payload.keys()) == 3)
        assert ('spans' in payload)
        assert (isinstance(payload['spans'], list))
        assert (len(payload['spans']) == 0)
        assert ('metrics' in payload)
        assert (len(payload['metrics'].keys()) == 1)
        assert ('plugins' in payload['metrics'])
        assert (isinstance(payload['metrics']['plugins'], list))
        assert (len(payload['metrics']['plugins']) == 1)

        python_plugin = payload['metrics']['plugins'][0]
        assert python_plugin['name'] == 'com.instana.plugin.python'
        assert python_plugin['entityId'] == str(os.getpid())
        assert 'data' in python_plugin
        assert 'snapshot' in python_plugin['data']
        assert 'metrics' not in python_plugin['data']



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2021

""" Module to assist in the data collection about the google cloud run service revision instance entity """
import os

from ....log import logger
from instana.collector.helpers.base import BaseHelper
from ....util import DictionaryOfStan


class InstanceEntityHelper(BaseHelper):
    """ This class helps in collecting data about the google cloud run service revision instance entity """

    def collect_metrics(self, **kwargs):
        """
        Collect and return metrics data (and optionally snapshot data) for this task
        @return: list - with one plugin entity
        """
        plugins = []
        plugin_data = dict()
        instance_metadata = kwargs.get('instance_metadata', {})
        project_metadata = kwargs.get('project_metadata', {})
        try:
            plugin_data["name"] = "com.instana.plugin.gcp.run.revision.instance"
            plugin_data["entityId"] = instance_metadata.get("id")
            plugin_data["data"] = DictionaryOfStan()
            plugin_data["data"]["runtime"] = "python"
            plugin_data["data"]["region"] = instance_metadata.get("region").split("/")[-1]
            plugin_data["data"]["service"] = self.collector.service
            plugin_data["data"]["configuration"] = self.collector.configuration
            plugin_data["data"]["revision"] = self.collector.revision
            plugin_data["data"]["instanceId"] = plugin_data["entityId"]
            plugin_data["data"]["port"] = os.getenv("PORT", "")
            plugin_data["data"]["numericProjectId"] = project_metadata.get("numericProjectId")
            plugin_data["data"]["projectId"] = project_metadata.get("projectId")

        except Exception:
            logger.debug("collect_service_revision_entity_metrics: ", exc_info=True)
        finally:
            plugins.append(plugin_data)

        return plugins



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2021

from instana.propagators.http_propagator import HTTPPropagator
from instana.w3c_trace_context.traceparent import Traceparent
from instana.span_context import SpanContext
from mock import patch
import unittest


class TestHTTPPropagatorTC(unittest.TestCase):
    def setUp(self):
        self.hptc = HTTPPropagator()

    @patch.object(Traceparent, "get_traceparent_fields")
    @patch.object(Traceparent, "validate")
    def test_extract_carrier_dict(self, mock_validate, mock_get_traceparent_fields):
        carrier = {
            'traceparent': '00-4bf92f3577b34da6a3ce929d0e0e4736-00f067aa0ba902b7-01',
            'tracestate': 'congo=t61rcWkgMzE',
            'X-INSTANA-T': '1234d0e0e4736234',
            'X-INSTANA-S': '1234567890abcdef',
            'X-INSTANA-L': '1, correlationType=web; correlationId=1234567890abcdef'
        }
        mock_validate.return_value = '00-4bf92f3577b34da6a3ce929d0e0e4736-00f067aa0ba902b7-01'
        mock_get_traceparent_fields.return_value = ["00", "4bf92f3577b34da6a3ce929d0e0e4736", "00f067aa0ba902b7", "01"]
        ctx = self.hptc.extract(carrier)
        self.assertEqual(ctx.correlation_id, '1234567890abcdef')
        self.assertEqual(ctx.correlation_type, "web")
        self.assertIsNone(ctx.instana_ancestor)
        self.assertEqual(ctx.level, 1)
        self.assertEqual(ctx.long_trace_id, "4bf92f3577b34da6a3ce929d0e0e4736")
        self.assertEqual(ctx.span_id, "00f067aa0ba902b7")
        self.assertFalse(ctx.synthetic)
        self.assertEqual(ctx.trace_id, "a3ce929d0e0e4736")  # 16 last chars from traceparent trace_id
        self.assertTrue(ctx.trace_parent)
        self.assertEqual(ctx.traceparent, '00-4bf92f3577b34da6a3ce929d0e0e4736-00f067aa0ba902b7-01')
        self.assertEqual(ctx.tracestate, 'congo=t61rcWkgMzE')

    @patch.object(Traceparent, "get_traceparent_fields")
    @patch.object(Traceparent, "validate")
    def test_extract_carrier_list(self, mock_validate, mock_get_traceparent_fields):
        carrier = [('user-agent', 'python-requests/2.23.0'), ('accept-encoding', 'gzip, deflate'),
                   ('accept', '*/*'), ('connection', 'keep-alive'),
                   ('traceparent', '00-4bf92f3577b34da6a3ce929d0e0e4736-00f067aa0ba902b7-01'),
                   ('tracestate', 'congo=t61rcWkgMzE'),
                   ('X-INSTANA-T', '1234d0e0e4736234'),
                   ('X-INSTANA-S', '1234567890abcdef'),
                   ('X-INSTANA-L', '1')]

        mock_validate.return_value = '00-4bf92f3577b34da6a3ce929d0e0e4736-00f067aa0ba902b7-01'
        mock_get_traceparent_fields.return_value = ["00", "4bf92f3577b34da6a3ce929d0e0e4736", "00f067aa0ba902b7", "01"]
        ctx = self.hptc.extract(carrier)
        self.assertIsNone(ctx.correlation_id)
        self.assertIsNone(ctx.correlation_type)
        self.assertIsNone(ctx.instana_ancestor)
        self.assertEqual(ctx.level, 1)
        self.assertIsNone(ctx.long_trace_id)
        self.assertEqual(ctx.span_id, "1234567890abcdef")
        self.assertFalse(ctx.synthetic)
        self.assertEqual(ctx.trace_id, "1234d0e0e4736234")  # 16 last chars from traceparent trace_id
        self.assertIsNone(ctx.trace_parent)
        self.assertEqual(ctx.traceparent, '00-4bf92f3577b34da6a3ce929d0e0e4736-00f067aa0ba902b7-01')
        self.assertEqual(ctx.tracestate, 'congo=t61rcWkgMzE')

    @patch.object(Traceparent, "validate")
    def test_extract_carrier_dict_validate_Exception_None_returned(self, mock_validate):
        """
        In this test case the traceparent header fails the validation, so traceparent and tracestate are not gonna used
        Additionally because in the instana L header the correlation flags are present we need to start a new ctx and
        the present values of 'X-INSTANA-T', 'X-INSTANA-S' headers should no be used. This means the ctx should be None
        :param mock_validate:
        :return:
        """
        carrier = {
            'traceparent': '00-4bf92f3577b34da6a3ce929d0e0e4736-00f067aa0ba902b7-01',
            'tracestate': 'congo=t61rcWkgMzE',
            'X-INSTANA-T': '1234d0e0e4736234',
            'X-INSTANA-S': '1234567890abcdef',
            'X-INSTANA-L': '1, correlationType=web; correlationId=1234567890abcdef'
        }
        mock_validate.return_value = None
        ctx = self.hptc.extract(carrier)
        self.assertTrue(isinstance(ctx, SpanContext))
        assert ctx.trace_id is None
        assert ctx.span_id is None
        assert ctx.synthetic is False
        self.assertEqual(ctx.correlation_id, "1234567890abcdef")
        self.assertEqual(ctx.correlation_type, "web")

    @patch.object(Traceparent, "validate")
    def test_extract_fake_exception(self, mock_validate):
        carrier = {
            'traceparent': '00-4bf92f3577b34da6a3ce929d0e0e4736-00f067aa0ba902b7-01',
            'tracestate': 'congo=t61rcWkgMzE',
            'X-INSTANA-T': '1234d0e0e4736234',
            'X-INSTANA-S': '1234567890abcdef',
            'X-INSTANA-L': '1, correlationType=web; correlationId=1234567890abcdef'
        }
        mock_validate.side_effect = Exception
        ctx = self.hptc.extract(carrier)
        self.assertIsNone(ctx)

    @patch.object(Traceparent, "get_traceparent_fields")
    @patch.object(Traceparent, "validate")
    def test_extract_carrier_dict_corrupted_level_header(self, mock_validate, mock_get_traceparent_fields):
        """
        In this test case the traceparent header fails the validation, so traceparent and tracestate are not gonna used
        Additionally because in the instana L header the correlation flags are present we need to start a new ctx and
        the present values of 'X-INSTANA-T', 'X-INSTANA-S' headers should no be used. This means the ctx should be None
        :param mock_validate:
        :return:
        """
        carrier = {
            'traceparent': '00-4bf92f3577b34da6a3ce929d0e0e4736-00f067aa0ba902b7-01',
            'tracestate': 'congo=t61rcWkgMzE',
            'X-INSTANA-T': '1234d0e0e4736234',
            'X-INSTANA-S': '1234567890abcdef',
            'X-INSTANA-L': '1, correlationTypeweb; correlationId1234567890abcdef'
        }
        mock_validate.return_value = '00-4bf92f3577b34da6a3ce929d0e0e4736-00f067aa0ba902b7-01'
        mock_get_traceparent_fields.return_value = ["00", "4bf92f3577b34da6a3ce929d0e0e4736", "00f067aa0ba902b7", "01"]
        ctx = self.hptc.extract(carrier)
        self.assertIsNone(ctx.correlation_id)
        self.assertIsNone(ctx.correlation_type)
        self.assertIsNone(ctx.instana_ancestor)
        self.assertEqual(ctx.level, 1)
        self.assertEqual(ctx.long_trace_id, '4bf92f3577b34da6a3ce929d0e0e4736')
        self.assertEqual(ctx.span_id, "00f067aa0ba902b7")
        self.assertFalse(ctx.synthetic)
        self.assertEqual(ctx.trace_id, "a3ce929d0e0e4736")  # 16 last chars from traceparent trace_id
        self.assertTrue(ctx.trace_parent)
        self.assertEqual(ctx.traceparent, '00-4bf92f3577b34da6a3ce929d0e0e4736-00f067aa0ba902b7-01')
        self.assertEqual(ctx.tracestate, 'congo=t61rcWkgMzE')

    @patch.object(Traceparent, "get_traceparent_fields")
    @patch.object(Traceparent, "validate")
    def test_extract_carrier_dict_level_header_not_splitable(self, mock_validate, mock_get_traceparent_fields):
        """
        In this test case the traceparent header fails the validation, so traceparent and tracestate are not gonna used
        Additionally because in the instana L header the correlation flags are present we need to start a new ctx and
        the present values of 'X-INSTANA-T', 'X-INSTANA-S' headers should no be used. This means the ctx should be None
        :param mock_validate:
        :return:
        """
        carrier = {
            'traceparent': '00-4bf92f3577b34da6a3ce929d0e0e4736-00f067aa0ba902b7-01',
            'tracestate': 'congo=t61rcWkgMzE',
            'X-INSTANA-T': '1234d0e0e4736234',
            'X-INSTANA-S': '1234567890abcdef',
            'X-INSTANA-L': ['1']
        }
        mock_validate.return_value = '00-4bf92f3577b34da6a3ce929d0e0e4736-00f067aa0ba902b7-01'
        mock_get_traceparent_fields.return_value = ["00", "4bf92f3577b34da6a3ce929d0e0e4736", "00f067aa0ba902b7", "01"]
        ctx = self.hptc.extract(carrier)
        self.assertIsNone(ctx.correlation_id)
        self.assertIsNone(ctx.correlation_type)
        self.assertIsNone(ctx.instana_ancestor)
        self.assertEqual(ctx.level, 1)
        self.assertIsNone(ctx.long_trace_id)
        self.assertEqual(ctx.span_id, "1234567890abcdef")
        self.assertFalse(ctx.synthetic)
        self.assertEqual(ctx.trace_id, "1234d0e0e4736234")
        self.assertIsNone(ctx.trace_parent)
        self.assertEqual(ctx.traceparent, '00-4bf92f3577b34da6a3ce929d0e0e4736-00f067aa0ba902b7-01')
        self.assertEqual(ctx.tracestate, 'congo=t61rcWkgMzE')


# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

"""
Instrumentation for Starlette
https://www.starlette.io/
"""
try:
    import starlette
    import wrapt
    from ..log import logger
    from .asgi import InstanaASGIMiddleware
    from starlette.middleware import Middleware

    @wrapt.patch_function_wrapper('starlette.applications', 'Starlette.__init__')
    def init_with_instana(wrapped, instance, args, kwargs):
        middleware = kwargs.get('middleware')
        if middleware is None:
            kwargs['middleware'] = [Middleware(InstanaASGIMiddleware)]
        elif isinstance(middleware, list):
            middleware.append(Middleware(InstanaASGIMiddleware))

        return wrapped(*args, **kwargs)

    logger.debug("Instrumenting Starlette")
except ImportError:
    pass



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

from __future__ import absolute_import

from pyramid.httpexceptions import HTTPException

import opentracing as ot
import opentracing.ext.tags as ext

from ...log import logger
from ...singletons import tracer, agent
from ...util.secrets import strip_secrets_from_query


class InstanaTweenFactory(object):
    """A factory that provides Instana instrumentation tween for Pyramid apps"""

    def __init__(self, handler, registry):
        self.handler = handler

    def __call__(self, request):
        ctx = tracer.extract(ot.Format.HTTP_HEADERS, dict(request.headers))
        scope = tracer.start_active_span('http', child_of=ctx)

        scope.span.set_tag(ext.SPAN_KIND, ext.SPAN_KIND_RPC_SERVER)
        scope.span.set_tag("http.host", request.host)
        scope.span.set_tag(ext.HTTP_METHOD, request.method)
        scope.span.set_tag(ext.HTTP_URL, request.path)

        if request.matched_route is not None:
            scope.span.set_tag("http.path_tpl", request.matched_route.pattern)

        if agent.options.extra_http_headers is not None:
            for custom_header in agent.options.extra_http_headers:
                # Headers are available in this format: HTTP_X_CAPTURE_THIS
                h = ('HTTP_' + custom_header.upper()).replace('-', '_')
                if h in request.headers:
                    scope.span.set_tag("http.header.%s" % custom_header, request.headers[h])

        if len(request.query_string):
            scrubbed_params = strip_secrets_from_query(request.query_string, agent.options.secrets_matcher,
                                                       agent.options.secrets_list)
            scope.span.set_tag("http.params", scrubbed_params)

        response = None
        try:
            response = self.handler(request)

            tracer.inject(scope.span.context, ot.Format.HTTP_HEADERS, response.headers)
            response.headers['Server-Timing'] = "intid;desc=%s" % scope.span.context.trace_id
        except HTTPException as e:
            response = e
            raise
        except BaseException as e:
            scope.span.set_tag("http.status", 500)

            # we need to explicitly populate the `message` tag with an error here
            # so that it's picked up from an SDK span
            scope.span.set_tag("message", str(e))
            scope.span.log_exception(e)

            logger.debug("Pyramid Instana tween", exc_info=True)
        finally:
            if response:
                scope.span.set_tag("http.status", response.status_int)

                if 500 <= response.status_int <= 511:
                    if response.exception is not None:
                        message = str(response.exception)
                        scope.span.log_exception(response.exception)
                    else:
                        message = response.status

                    scope.span.set_tag("message", message)
                    scope.span.assure_errored()

            scope.close()

        return response


def includeme(config):
    logger.debug("Instrumenting pyramid")
    config.add_tween(__name__ + '.InstanaTweenFactory')



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

from __future__ import absolute_import

import sys
import unittest
import pytest
import json
import requests
import io

from instana.singletons import tracer
from ..test_utils import _TraceContextMixin

from mock import patch, Mock
from six.moves import http_client

if sys.version_info[0] >= 3:
    from google.cloud import storage
    from google.api_core import iam

@pytest.mark.skipif(sys.version_info[0] < 3, reason="google-cloud-storage has dropped support for Python 2")
class TestGoogleCloudStorage(unittest.TestCase, _TraceContextMixin):
    def setUp(self):
        self.recorder = tracer.recorder
        self.recorder.clear_spans()

    @patch('requests.Session.request')
    def test_buckets_list(self, mock_requests):
        mock_requests.return_value = self._mock_response(
            json_content={"kind": "storage#buckets", "items": []},
            status_code=http_client.OK
        )

        client = self._client(project='test-project')

        with tracer.start_active_span('test'):
            buckets = client.list_buckets()
            self.assertEqual(0, self.recorder.queue_size(), msg='span has been created before the actual request')

            # trigger the iterator
            for b in buckets:
                pass

        spans = self.recorder.queued_spans()

        self.assertEqual(2, len(spans))
        self.assertIsNone(tracer.active_span)

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        self.assertEqual('gcs',gcs_span.n)
        self.assertEqual(2, gcs_span.k)
        self.assertIsNone(gcs_span.ec)

        self.assertEqual('buckets.list', gcs_span.data["gcs"]["op"])
        self.assertEqual('test-project', gcs_span.data["gcs"]["projectId"])

    @patch('requests.Session.request')
    def test_buckets_insert(self, mock_requests):
        mock_requests.return_value = self._mock_response(
            json_content={"kind": "storage#bucket"},
            status_code=http_client.OK
        )

        client = self._client(project='test-project')

        with tracer.start_active_span('test'):
            client.create_bucket('test bucket')

        spans = self.recorder.queued_spans()

        self.assertEqual(2, len(spans))
        self.assertIsNone(tracer.active_span)

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        self.assertEqual('gcs',gcs_span.n)
        self.assertEqual(2, gcs_span.k)
        self.assertIsNone(gcs_span.ec)

        self.assertEqual('buckets.insert', gcs_span.data["gcs"]["op"])
        self.assertEqual('test-project', gcs_span.data["gcs"]["projectId"])
        self.assertEqual('test bucket', gcs_span.data["gcs"]["bucket"])

    @patch('requests.Session.request')
    def test_buckets_get(self, mock_requests):
        mock_requests.return_value = self._mock_response(
            json_content={"kind": "storage#bucket"},
            status_code=http_client.OK
        )

        client = self._client(project='test-project')

        with tracer.start_active_span('test'):
            client.get_bucket('test bucket')

        spans = self.recorder.queued_spans()

        self.assertEqual(2, len(spans))
        self.assertIsNone(tracer.active_span)

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertEqual(test_span.t, gcs_span.t)
        self.assertEqual(test_span.s, gcs_span.p)

        self.assertEqual('gcs',gcs_span.n)
        self.assertEqual(2, gcs_span.k)
        self.assertIsNone(gcs_span.ec)

        self.assertEqual('buckets.get', gcs_span.data["gcs"]["op"])
        self.assertEqual('test bucket', gcs_span.data["gcs"]["bucket"])

    @patch('requests.Session.request')
    def test_buckets_patch(self, mock_requests):
        mock_requests.return_value = self._mock_response(
            json_content={"kind": "storage#bucket"},
            status_code=http_client.OK
        )

        client = self._client(project='test-project')

        with tracer.start_active_span('test'):
            client.bucket('test bucket').patch()

        spans = self.recorder.queued_spans()

        self.assertEqual(2, len(spans))
        self.assertIsNone(tracer.active_span)

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        self.assertEqual('gcs',gcs_span.n)
        self.assertEqual(2, gcs_span.k)
        self.assertIsNone(gcs_span.ec)

        self.assertEqual('buckets.patch', gcs_span.data["gcs"]["op"])
        self.assertEqual('test bucket', gcs_span.data["gcs"]["bucket"])

    @patch('requests.Session.request')
    def test_buckets_update(self, mock_requests):
        mock_requests.return_value = self._mock_response(
            json_content={"kind": "storage#bucket"},
            status_code=http_client.OK
        )

        client = self._client(project='test-project')

        with tracer.start_active_span('test'):
            client.bucket('test bucket').update()

        spans = self.recorder.queued_spans()

        self.assertEqual(2, len(spans))
        self.assertIsNone(tracer.active_span)

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        self.assertEqual('gcs',gcs_span.n)
        self.assertEqual(2, gcs_span.k)
        self.assertIsNone(gcs_span.ec)

        self.assertEqual('buckets.update', gcs_span.data["gcs"]["op"])
        self.assertEqual('test bucket', gcs_span.data["gcs"]["bucket"])

    @patch('requests.Session.request')
    def test_buckets_get_iam_policy(self, mock_requests):
        mock_requests.return_value = self._mock_response(
            json_content={"kind": "storage#policy"},
            status_code=http_client.OK
        )

        client = self._client(project='test-project')

        with tracer.start_active_span('test'):
            client.bucket('test bucket').get_iam_policy()

        spans = self.recorder.queued_spans()

        self.assertEqual(2, len(spans))
        self.assertIsNone(tracer.active_span)

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        self.assertEqual('gcs',gcs_span.n)
        self.assertEqual(2, gcs_span.k)
        self.assertIsNone(gcs_span.ec)

        self.assertEqual('buckets.getIamPolicy', gcs_span.data["gcs"]["op"])
        self.assertEqual('test bucket', gcs_span.data["gcs"]["bucket"])

    @patch('requests.Session.request')
    def test_buckets_set_iam_policy(self, mock_requests):
        mock_requests.return_value = self._mock_response(
            json_content={"kind": "storage#policy"},
            status_code=http_client.OK
        )

        client = self._client(project='test-project')

        with tracer.start_active_span('test'):
            client.bucket('test bucket').set_iam_policy(iam.Policy())

        spans = self.recorder.queued_spans()

        self.assertEqual(2, len(spans))
        self.assertIsNone(tracer.active_span)

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        self.assertEqual('gcs',gcs_span.n)
        self.assertEqual(2, gcs_span.k)
        self.assertIsNone(gcs_span.ec)

        self.assertEqual('buckets.setIamPolicy', gcs_span.data["gcs"]["op"])
        self.assertEqual('test bucket', gcs_span.data["gcs"]["bucket"])

    @patch('requests.Session.request')
    def test_buckets_test_iam_permissions(self, mock_requests):
        mock_requests.return_value = self._mock_response(
            json_content={"kind": "storage#testIamPermissionsResponse"},
            status_code=http_client.OK
        )

        client = self._client(project='test-project')

        with tracer.start_active_span('test'):
            client.bucket('test bucket').test_iam_permissions('test-permission')

        spans = self.recorder.queued_spans()

        self.assertEqual(2, len(spans))
        self.assertIsNone(tracer.active_span)

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        self.assertEqual('gcs',gcs_span.n)
        self.assertEqual(2, gcs_span.k)
        self.assertIsNone(gcs_span.ec)

        self.assertEqual('buckets.testIamPermissions', gcs_span.data["gcs"]["op"])
        self.assertEqual('test bucket', gcs_span.data["gcs"]["bucket"])

    @patch('requests.Session.request')
    def test_buckets_lock_retention_policy(self, mock_requests):
        mock_requests.return_value = self._mock_response(
            json_content={"kind": "storage#bucket", "metageneration": 1, "retentionPolicy": {"isLocked": False}},
            status_code=http_client.OK
        )

        client = self._client(project='test-project')

        bucket = client.bucket('test bucket')
        bucket.reload()

        with tracer.start_active_span('test'):
            bucket.lock_retention_policy()

        spans = self.recorder.queued_spans()

        self.assertEqual(2, len(spans))
        self.assertIsNone(tracer.active_span)

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        self.assertEqual('gcs',gcs_span.n)
        self.assertEqual(2, gcs_span.k)
        self.assertIsNone(gcs_span.ec)

        self.assertEqual('buckets.lockRetentionPolicy', gcs_span.data["gcs"]["op"])
        self.assertEqual('test bucket', gcs_span.data["gcs"]["bucket"])

    @patch('requests.Session.request')
    def test_buckets_delete(self, mock_requests):
        mock_requests.return_value = self._mock_response()

        client = self._client(project='test-project')

        with tracer.start_active_span('test'):
            client.bucket('test bucket').delete()

        spans = self.recorder.queued_spans()

        self.assertEqual(2, len(spans))
        self.assertIsNone(tracer.active_span)

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        self.assertEqual('gcs',gcs_span.n)
        self.assertEqual(2, gcs_span.k)
        self.assertIsNone(gcs_span.ec)

        self.assertEqual('buckets.delete', gcs_span.data["gcs"]["op"])
        self.assertEqual('test bucket', gcs_span.data["gcs"]["bucket"])

    @patch('requests.Session.request')
    def test_objects_compose(self, mock_requests):
        mock_requests.return_value = self._mock_response(
            json_content={"kind": "storage#object"},
            status_code=http_client.OK
        )

        client = self._client(project='test-project')

        with tracer.start_active_span('test'):
            client.bucket('test bucket').blob('dest object').compose([
                storage.blob.Blob('object 1', 'test bucket'),
                storage.blob.Blob('object 2', 'test bucket')
            ])

        spans = self.recorder.queued_spans()

        self.assertEqual(2, len(spans))
        self.assertIsNone(tracer.active_span)

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        self.assertEqual('gcs',gcs_span.n)
        self.assertEqual(2, gcs_span.k)
        self.assertIsNone(gcs_span.ec)

        self.assertEqual('objects.compose', gcs_span.data["gcs"]["op"])
        self.assertEqual('test bucket', gcs_span.data["gcs"]["destinationBucket"])
        self.assertEqual('dest object', gcs_span.data["gcs"]["destinationObject"])
        self.assertEqual('test bucket/object 1,test bucket/object 2', gcs_span.data["gcs"]["sourceObjects"])

    @patch('requests.Session.request')
    def test_objects_copy(self, mock_requests):
        mock_requests.return_value = self._mock_response(
            json_content={"kind": "storage#object"},
            status_code=http_client.OK
        )

        client = self._client(project='test-project')
        bucket = client.bucket('src bucket')

        with tracer.start_active_span('test'):
            bucket.copy_blob(
                bucket.blob('src object'),
                client.bucket('dest bucket'),
                new_name='dest object'
            )

        spans = self.recorder.queued_spans()

        self.assertEqual(2, len(spans))
        self.assertIsNone(tracer.active_span)

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        self.assertEqual('gcs',gcs_span.n)
        self.assertEqual(2, gcs_span.k)
        self.assertIsNone(gcs_span.ec)

        self.assertEqual('objects.copy', gcs_span.data["gcs"]["op"])
        self.assertEqual('dest bucket', gcs_span.data["gcs"]["destinationBucket"])
        self.assertEqual('dest object', gcs_span.data["gcs"]["destinationObject"])
        self.assertEqual('src bucket', gcs_span.data["gcs"]["sourceBucket"])
        self.assertEqual('src object', gcs_span.data["gcs"]["sourceObject"])

    @patch('requests.Session.request')
    def test_objects_delete(self, mock_requests):
        mock_requests.return_value = self._mock_response()

        client = self._client(project='test-project')

        with tracer.start_active_span('test'):
            client.bucket('test bucket').blob('test object').delete()

        spans = self.recorder.queued_spans()

        self.assertEqual(2, len(spans))
        self.assertIsNone(tracer.active_span)

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        self.assertEqual('gcs',gcs_span.n)
        self.assertEqual(2, gcs_span.k)
        self.assertIsNone(gcs_span.ec)

        self.assertEqual('objects.delete', gcs_span.data["gcs"]["op"])
        self.assertEqual('test bucket', gcs_span.data["gcs"]["bucket"])
        self.assertEqual('test object', gcs_span.data["gcs"]["object"])

    @patch('requests.Session.request')
    def test_objects_attrs(self, mock_requests):
        mock_requests.return_value = self._mock_response(
            json_content={"kind": "storage#object"},
            status_code=http_client.OK
        )

        client = self._client(project='test-project')

        with tracer.start_active_span('test'):
            client.bucket('test bucket').blob('test object').exists()

        spans = self.recorder.queued_spans()

        self.assertEqual(2, len(spans))
        self.assertIsNone(tracer.active_span)

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        self.assertEqual('gcs',gcs_span.n)
        self.assertEqual(2, gcs_span.k)
        self.assertIsNone(gcs_span.ec)

        self.assertEqual('objects.attrs', gcs_span.data["gcs"]["op"])
        self.assertEqual('test bucket', gcs_span.data["gcs"]["bucket"])
        self.assertEqual('test object', gcs_span.data["gcs"]["object"])

    @patch('requests.Session.request')
    def test_objects_get(self, mock_requests):
        mock_requests.return_value = self._mock_response(
            content=b'CONTENT',
            status_code=http_client.OK
        )

        client = self._client(project='test-project')

        with tracer.start_active_span('test'):
            client.bucket('test bucket').blob('test object').download_to_file(
                io.BytesIO(),
                raw_download=True
            )

        spans = self.recorder.queued_spans()

        self.assertEqual(2, len(spans))
        self.assertIsNone(tracer.active_span)

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        self.assertEqual('gcs',gcs_span.n)
        self.assertEqual(2, gcs_span.k)
        self.assertIsNone(gcs_span.ec)

        self.assertEqual('objects.get', gcs_span.data["gcs"]["op"])
        self.assertEqual('test bucket', gcs_span.data["gcs"]["bucket"])
        self.assertEqual('test object', gcs_span.data["gcs"]["object"])

    @patch('requests.Session.request')
    def test_objects_insert(self, mock_requests):
        mock_requests.return_value = self._mock_response(
            json_content={"kind": "storage#object"},
            status_code=http_client.OK
        )

        client = self._client(project='test-project')

        with tracer.start_active_span('test'):
            client.bucket('test bucket').blob('test object').upload_from_string('CONTENT')

        spans = self.recorder.queued_spans()

        self.assertEqual(2, len(spans))
        self.assertIsNone(tracer.active_span)

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        self.assertEqual('gcs',gcs_span.n)
        self.assertEqual(2, gcs_span.k)
        self.assertIsNone(gcs_span.ec)

        self.assertEqual('objects.insert', gcs_span.data["gcs"]["op"])
        self.assertEqual('test bucket', gcs_span.data["gcs"]["bucket"])
        self.assertEqual('test object', gcs_span.data["gcs"]["object"])

    @patch('requests.Session.request')
    def test_objects_list(self, mock_requests):
        mock_requests.return_value = self._mock_response(
            json_content={"kind": "storage#object"},
            status_code=http_client.OK
        )

        client = self._client(project='test-project')

        with tracer.start_active_span('test'):
            blobs = client.bucket('test bucket').list_blobs()
            self.assertEqual(0, self.recorder.queue_size(), msg='span has been created before the actual request')

            for b in blobs: pass

        spans = self.recorder.queued_spans()

        self.assertEqual(2, len(spans))
        self.assertIsNone(tracer.active_span)

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        self.assertEqual('gcs',gcs_span.n)
        self.assertEqual(2, gcs_span.k)
        self.assertIsNone(gcs_span.ec)

        self.assertEqual('objects.list', gcs_span.data["gcs"]["op"])
        self.assertEqual('test bucket', gcs_span.data["gcs"]["bucket"])

    @patch('requests.Session.request')
    def test_objects_patch(self, mock_requests):
        mock_requests.return_value = self._mock_response(
            json_content={"kind": "storage#object"},
            status_code=http_client.OK
        )

        client = self._client(project='test-project')

        with tracer.start_active_span('test'):
            client.bucket('test bucket').blob('test object').patch()

        spans = self.recorder.queued_spans()

        self.assertEqual(2, len(spans))
        self.assertIsNone(tracer.active_span)

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        self.assertEqual('gcs',gcs_span.n)
        self.assertEqual(2, gcs_span.k)
        self.assertIsNone(gcs_span.ec)

        self.assertEqual('objects.patch', gcs_span.data["gcs"]["op"])
        self.assertEqual('test bucket', gcs_span.data["gcs"]["bucket"])
        self.assertEqual('test object', gcs_span.data["gcs"]["object"])

    @patch('requests.Session.request')
    def test_objects_rewrite(self, mock_requests):
        mock_requests.return_value = self._mock_response(
            json_content={"kind": "storage#rewriteResponse", "totalBytesRewritten": 0, "objectSize": 0, "done": True, "resource": {}},
            status_code=http_client.OK
        )

        client = self._client(project='test-project')

        with tracer.start_active_span('test'):
            client.bucket('dest bucket').blob('dest object').rewrite(
                client.bucket('src bucket').blob('src object')
            )

        spans = self.recorder.queued_spans()

        self.assertEqual(2, len(spans))
        self.assertIsNone(tracer.active_span)

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        self.assertEqual('gcs',gcs_span.n)
        self.assertEqual(2, gcs_span.k)
        self.assertIsNone(gcs_span.ec)

        self.assertEqual('objects.rewrite', gcs_span.data["gcs"]["op"])
        self.assertEqual('dest bucket', gcs_span.data["gcs"]["destinationBucket"])
        self.assertEqual('dest object', gcs_span.data["gcs"]["destinationObject"])
        self.assertEqual('src bucket', gcs_span.data["gcs"]["sourceBucket"])
        self.assertEqual('src object', gcs_span.data["gcs"]["sourceObject"])

    @patch('requests.Session.request')
    def test_objects_update(self, mock_requests):
        mock_requests.return_value = self._mock_response(
            json_content={"kind": "storage#object"},
            status_code=http_client.OK
        )

        client = self._client(project='test-project')

        with tracer.start_active_span('test'):
            client.bucket('test bucket').blob('test object').update()

        spans = self.recorder.queued_spans()

        self.assertEqual(2, len(spans))
        self.assertIsNone(tracer.active_span)

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        self.assertEqual('gcs',gcs_span.n)
        self.assertEqual(2, gcs_span.k)
        self.assertIsNone(gcs_span.ec)

        self.assertEqual('objects.update', gcs_span.data["gcs"]["op"])
        self.assertEqual('test bucket', gcs_span.data["gcs"]["bucket"])
        self.assertEqual('test object', gcs_span.data["gcs"]["object"])

    @patch('requests.Session.request')
    def test_default_acls_list(self, mock_requests):
        mock_requests.return_value = self._mock_response(
            json_content={"kind": "storage#objectAccessControls", "items": []},
            status_code=http_client.OK
        )

        client = self._client(project='test-project')

        with tracer.start_active_span('test'):
            client.bucket('test bucket').default_object_acl.get_entities()

        spans = self.recorder.queued_spans()

        self.assertEqual(2, len(spans))
        self.assertIsNone(tracer.active_span)

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        self.assertEqual('gcs',gcs_span.n)
        self.assertEqual(2, gcs_span.k)
        self.assertIsNone(gcs_span.ec)

        self.assertEqual('defaultAcls.list', gcs_span.data["gcs"]["op"])
        self.assertEqual('test bucket', gcs_span.data["gcs"]["bucket"])

    @patch('requests.Session.request')
    def test_object_acls_list(self, mock_requests):
        mock_requests.return_value = self._mock_response(
            json_content={"kind": "storage#objectAccessControls", "items": []},
            status_code=http_client.OK
        )

        client = self._client(project='test-project')

        with tracer.start_active_span('test'):
            client.bucket('test bucket').blob('test object').acl.get_entities()

        spans = self.recorder.queued_spans()

        self.assertEqual(2, len(spans))
        self.assertIsNone(tracer.active_span)

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        self.assertEqual('gcs',gcs_span.n)
        self.assertEqual(2, gcs_span.k)
        self.assertIsNone(gcs_span.ec)

        self.assertEqual('objectAcls.list', gcs_span.data["gcs"]["op"])
        self.assertEqual('test bucket', gcs_span.data["gcs"]["bucket"])
        self.assertEqual('test object', gcs_span.data["gcs"]["object"])

    @patch('requests.Session.request')
    def test_object_hmac_keys_create(self, mock_requests):
        mock_requests.return_value = self._mock_response(
            json_content={"kind": "storage#hmacKey", "metadata": {}, "secret": ""},
            status_code=http_client.OK
        )

        client = self._client(project='test-project')

        with tracer.start_active_span('test'):
            client.create_hmac_key('test@example.com')

        spans = self.recorder.queued_spans()

        self.assertEqual(2, len(spans))
        self.assertIsNone(tracer.active_span)

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        self.assertEqual('gcs',gcs_span.n)
        self.assertEqual(2, gcs_span.k)
        self.assertIsNone(gcs_span.ec)

        self.assertEqual('hmacKeys.create', gcs_span.data["gcs"]["op"])
        self.assertEqual('test-project', gcs_span.data["gcs"]["projectId"])

    @patch('requests.Session.request')
    def test_object_hmac_keys_delete(self, mock_requests):
        mock_requests.return_value = self._mock_response()

        client = self._client(project='test-project')

        with tracer.start_active_span('test'):
            key = storage.hmac_key.HMACKeyMetadata(client, access_id='test key')
            key.state = storage.hmac_key.HMACKeyMetadata.INACTIVE_STATE
            key.delete()

        spans = self.recorder.queued_spans()

        self.assertEqual(2, len(spans))
        self.assertIsNone(tracer.active_span)

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        self.assertEqual('gcs',gcs_span.n)
        self.assertEqual(2, gcs_span.k)
        self.assertIsNone(gcs_span.ec)

        self.assertEqual('hmacKeys.delete', gcs_span.data["gcs"]["op"])
        self.assertEqual('test-project', gcs_span.data["gcs"]["projectId"])
        self.assertEqual('test key', gcs_span.data["gcs"]["accessId"])

    @patch('requests.Session.request')
    def test_object_hmac_keys_get(self, mock_requests):
        mock_requests.return_value = self._mock_response(
            json_content={"kind": "storage#hmacKey", "metadata": {}, "secret": ""},
            status_code=http_client.OK
        )

        client = self._client(project='test-project')

        with tracer.start_active_span('test'):
            storage.hmac_key.HMACKeyMetadata(client, access_id='test key').exists()

        spans = self.recorder.queued_spans()

        self.assertEqual(2, len(spans))
        self.assertIsNone(tracer.active_span)

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        self.assertEqual('gcs',gcs_span.n)
        self.assertEqual(2, gcs_span.k)
        self.assertIsNone(gcs_span.ec)

        self.assertEqual('hmacKeys.get', gcs_span.data["gcs"]["op"])
        self.assertEqual('test-project', gcs_span.data["gcs"]["projectId"])
        self.assertEqual('test key', gcs_span.data["gcs"]["accessId"])

    @patch('requests.Session.request')
    def test_object_hmac_keys_list(self, mock_requests):
        mock_requests.return_value = self._mock_response(
            json_content={"kind": "storage#hmacKeysMetadata", "items": []},
            status_code=http_client.OK
        )

        client = self._client(project='test-project')

        with tracer.start_active_span('test'):
            keys = client.list_hmac_keys()
            self.assertEqual(0, self.recorder.queue_size(), msg='span has been created before the actual request')

            for k in keys: pass

        spans = self.recorder.queued_spans()

        self.assertEqual(2, len(spans))
        self.assertIsNone(tracer.active_span)

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        self.assertEqual('gcs',gcs_span.n)
        self.assertEqual(2, gcs_span.k)
        self.assertIsNone(gcs_span.ec)

        self.assertEqual('hmacKeys.list', gcs_span.data["gcs"]["op"])
        self.assertEqual('test-project', gcs_span.data["gcs"]["projectId"])

    @patch('requests.Session.request')
    def test_object_hmac_keys_update(self, mock_requests):
        mock_requests.return_value = self._mock_response(
            json_content={"kind": "storage#hmacKey", "metadata": {}, "secret": ""},
            status_code=http_client.OK
        )

        client = self._client(project='test-project')

        with tracer.start_active_span('test'):
            storage.hmac_key.HMACKeyMetadata(client, access_id='test key').update()

        spans = self.recorder.queued_spans()

        self.assertEqual(2, len(spans))
        self.assertIsNone(tracer.active_span)

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        self.assertEqual('gcs',gcs_span.n)
        self.assertEqual(2, gcs_span.k)
        self.assertIsNone(gcs_span.ec)

        self.assertEqual('hmacKeys.update', gcs_span.data["gcs"]["op"])
        self.assertEqual('test-project', gcs_span.data["gcs"]["projectId"])
        self.assertEqual('test key', gcs_span.data["gcs"]["accessId"])

    @patch('requests.Session.request')
    def test_object_hmac_keys_update(self, mock_requests):
        mock_requests.return_value = self._mock_response(
            json_content={"email_address": "test@example.com", "kind": "storage#serviceAccount"},
            status_code=http_client.OK
        )

        client = self._client(project='test-project')

        with tracer.start_active_span('test'):
            client.get_service_account_email()

        spans = self.recorder.queued_spans()

        self.assertEqual(2, len(spans))
        self.assertIsNone(tracer.active_span)

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        self.assertEqual('gcs',gcs_span.n)
        self.assertEqual(2, gcs_span.k)
        self.assertIsNone(gcs_span.ec)

        self.assertEqual('serviceAccount.get', gcs_span.data["gcs"]["op"])
        self.assertEqual('test-project', gcs_span.data["gcs"]["projectId"])

    @patch('requests.Session.request')
    def test_batch_operation(self, mock_requests):
        mock_requests.return_value = self._mock_response(
            _TWO_PART_BATCH_RESPONSE,
            status_code=http_client.OK,
            headers={"content-type": 'multipart/mixed; boundary="DEADBEEF="'}
        )

        client = self._client(project='test-project')
        bucket = client.bucket('test-bucket')

        with tracer.start_active_span('test'):
            with client.batch():
                for obj in ['obj1', 'obj2']:
                    bucket.delete_blob(obj)

        spans = self.recorder.queued_spans()

        self.assertEqual(2, len(spans))

    def _client(self, *args, **kwargs):
        # override the HTTP client to bypass the authorization
        kwargs['_http'] = kwargs.get('_http', requests.Session())
        kwargs['_http'].is_mtls = False

        return storage.Client(*args, **kwargs)

    def _mock_response(self, content=b'', status_code=http_client.NO_CONTENT, json_content=None, headers={}):
        resp = Mock()
        resp.status_code = status_code
        resp.headers = headers
        resp.content = content
        resp.__enter__ = Mock(return_value=resp)
        resp.__exit__ = Mock()

        if json_content is not None:
            if resp.content == b'':
                resp.content = json.dumps(json_content)

            resp.json = Mock(return_value=json_content)

        return resp

_TWO_PART_BATCH_RESPONSE = b"""\
--DEADBEEF=
Content-Type: application/json
Content-ID: <response-1+1>

HTTP/1.1 204 No Content

Content-Type: application/json; charset=UTF-8
Content-Length: 0

--DEADBEEF=
Content-Type: application/json
Content-ID: <response-1+2>

HTTP/1.1 204 No Content

Content-Type: application/json; charset=UTF-8
Content-Length: 0

--DEADBEEF=--
"""



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2019

"""
couchbase instrumentation - This instrumentation supports the Python CouchBase 2.3.4 --> 2.5.x SDK currently:
https://docs.couchbase.com/python-sdk/2.5/start-using-sdk.html
"""
from __future__ import absolute_import

from distutils.version import LooseVersion
import wrapt

from ..log import logger
from ..util.traceutils import get_active_tracer

try:
    import couchbase
    from couchbase.n1ql import N1QLQuery

    # List of operations to instrument
    # incr, incr_multi, decr, decr_multi, retrieve_in are wrappers around operations above
    operations = ['upsert', 'insert', 'replace', 'append', 'prepend', 'get', 'rget',
                  'touch', 'lock', 'unlock', 'remove', 'counter', 'mutate_in', 'lookup_in',
                  'stats', 'ping', 'diagnostics', 'observe',

                  'upsert_multi', 'insert_multi', 'replace_multi', 'append_multi',
                  'prepend_multi', 'get_multi', 'touch_multi', 'lock_multi', 'unlock_multi',
                  'observe_multi', 'endure_multi', 'remove_multi', 'counter_multi']

    def capture_kvs(scope, instance, query_arg, op):
        try:
            scope.span.set_tag('couchbase.hostname', instance.server_nodes[0])
            scope.span.set_tag('couchbase.bucket', instance.bucket)
            scope.span.set_tag('couchbase.type', op)

            if query_arg is not None:
                query = None
                if type(query_arg) is N1QLQuery:
                    query = query_arg.statement
                else:
                    query = query_arg

                scope.span.set_tag('couchbase.sql', query)
        except:
            # No fail on key capture - best effort
            pass

    def make_wrapper(op):
        def wrapper(wrapped, instance, args, kwargs):
            active_tracer = get_active_tracer()

            # If we're not tracing, just return
            if active_tracer is None:
                return wrapped(*args, **kwargs)

            with active_tracer.start_active_span("couchbase", child_of=active_tracer.active_span) as scope:
                capture_kvs(scope, instance, None, op)
                try:
                    return wrapped(*args, **kwargs)
                except Exception as e:
                    scope.span.log_exception(e)
                    scope.span.set_tag('couchbase.error', repr(e))
                    raise
        return wrapper

    def query_with_instana(wrapped, instance, args, kwargs):
        active_tracer = get_active_tracer()

        # If we're not tracing, just return
        if active_tracer is None:
            return wrapped(*args, **kwargs)

        with active_tracer.start_active_span("couchbase", child_of=active_tracer.active_span) as scope:
            capture_kvs(scope, instance, args[0], 'n1ql_query')
            try:
                return wrapped(*args, **kwargs)
            except Exception as e:
                scope.span.log_exception(e)
                scope.span.set_tag('couchbase.error', repr(e))
                raise

    if hasattr(couchbase, '__version__') \
            and (LooseVersion(couchbase.__version__) >= LooseVersion('2.3.4')) \
            and (LooseVersion(couchbase.__version__) < LooseVersion('3.0.0')):
        logger.debug("Instrumenting couchbase")
        wrapt.wrap_function_wrapper('couchbase.bucket', 'Bucket.n1ql_query', query_with_instana)
        for op in operations:
            f = make_wrapper(op)
            wrapt.wrap_function_wrapper('couchbase.bucket', 'Bucket.%s' % op, f)

except ImportError:
    pass


# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import sys
from .app import soapserver as server
from ..utils import launch_background_thread

app_thread = None

if sys.version_info < (3, 7, 0) and app_thread is None:
    app_thread = launch_background_thread(server.serve_forever, "SoapServer")




# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

from __future__ import absolute_import

import opentracing
from ...log import logger
from ...singletons import tracer
from ...util.traceutils import get_active_tracer

try:
    import celery
    from celery import registry, signals
    from .catalog import task_catalog_get, task_catalog_pop, task_catalog_push, get_task_id

    try:
        from urllib import parse
    except ImportError:
        import urlparse as parse
        import urllib


    def add_broker_tags(span, broker_url):
        try:
            url = parse.urlparse(broker_url)

            # Add safety for edge case where scheme may not be a string
            url_scheme = str(url.scheme)
            span.set_tag("scheme", url_scheme)

            if url.hostname is None:
                span.set_tag("host", 'localhost')
            else:
                span.set_tag("host", url.hostname)

            if url.port is None:
                # Set default port if not specified
                if url_scheme == 'redis':
                    span.set_tag("port", "6379")
                elif 'amqp' in url_scheme:
                    span.set_tag("port", "5672")
                elif 'sqs' in url_scheme:
                    span.set_tag("port", "443")
            else:
                span.set_tag("port", str(url.port))
        except Exception:
            logger.debug("Error parsing broker URL: %s" % broker_url, exc_info=True)


    @signals.task_prerun.connect
    def task_prerun(*args, **kwargs):
        try:
            ctx = None
            task = kwargs.get('sender', None)
            task_id = kwargs.get('task_id', None)
            task = registry.tasks.get(task.name)

            headers = task.request.get('headers', {})
            if headers is not None:
                ctx = tracer.extract(opentracing.Format.HTTP_HEADERS, headers, disable_w3c_trace_context=True)

            scope = tracer.start_active_span("celery-worker", child_of=ctx)
            scope.span.set_tag("task", task.name)
            scope.span.set_tag("task_id", task_id)
            add_broker_tags(scope.span, task.app.conf['broker_url'])

            # Store the scope on the task to eventually close it out on the "after" signal
            task_catalog_push(task, task_id, scope, True)
        except:
            logger.debug("task_prerun: ", exc_info=True)


    @signals.task_postrun.connect
    def task_postrun(*args, **kwargs):
        try:
            task = kwargs.get('sender', None)
            task_id = kwargs.get('task_id', None)
            scope = task_catalog_pop(task, task_id, True)
            if scope is not None:
                scope.close()
        except:
            logger.debug("after_task_publish: ", exc_info=True)


    @signals.task_failure.connect
    def task_failure(*args, **kwargs):
        try:
            task_id = kwargs.get('task_id', None)
            task = kwargs['sender']
            scope = task_catalog_get(task, task_id, True)

            if scope is not None:
                scope.span.set_tag("success", False)
                exc = kwargs.get('exception', None)
                if exc is None:
                    scope.span.mark_as_errored()
                else:
                    scope.span.log_exception(kwargs['exception'])
        except:
            logger.debug("task_failure: ", exc_info=True)


    @signals.task_retry.connect
    def task_retry(*args, **kwargs):
        try:
            task_id = kwargs.get('task_id', None)
            task = kwargs['sender']
            scope = task_catalog_get(task, task_id, True)

            if scope is not None:
                reason = kwargs.get('reason', None)
                if reason is not None:
                    scope.span.set_tag('retry-reason', reason)
        except:
            logger.debug("task_failure: ", exc_info=True)


    @signals.before_task_publish.connect
    def before_task_publish(*args, **kwargs):
        try:
            active_tracer = get_active_tracer()
            if active_tracer is not None:
                body = kwargs['body']
                headers = kwargs['headers']
                task_name = kwargs['sender']
                task = registry.tasks.get(task_name)
                task_id = get_task_id(headers, body)

                scope = active_tracer.start_active_span("celery-client", child_of=active_tracer.active_span)
                scope.span.set_tag("task", task_name)
                scope.span.set_tag("task_id", task_id)
                add_broker_tags(scope.span, task.app.conf['broker_url'])

                # Context propagation
                context_headers = {}
                active_tracer.inject(scope.span.context, opentracing.Format.HTTP_HEADERS, context_headers,
                                     disable_w3c_trace_context=True)

                # Fix for broken header propagation
                # https://github.com/celery/celery/issues/4875
                task_headers = kwargs.get('headers') or {}
                task_headers.setdefault('headers', {})
                task_headers['headers'].update(context_headers)
                kwargs['headers'] = task_headers

                # Store the scope on the task to eventually close it out on the "after" signal
                task_catalog_push(task, task_id, scope, False)
        except:
            logger.debug("before_task_publish: ", exc_info=True)


    @signals.after_task_publish.connect
    def after_task_publish(*args, **kwargs):
        try:
            task_id = get_task_id(kwargs['headers'], kwargs['body'])
            task = registry.tasks.get(kwargs['sender'])
            scope = task_catalog_pop(task, task_id, False)
            if scope is not None:
                scope.close()
        except:
            logger.debug("after_task_publish: ", exc_info=True)


    logger.debug("Instrumenting celery")
except ImportError:
    pass



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

from __future__ import absolute_import

import asyncio
import aiohttp
import unittest

import tests.apps.flask_app
from ..helpers import testenv
from instana.configurator import config
from instana.singletons import async_tracer


class TestAsyncio(unittest.TestCase):
    def setUp(self):
        """ Clear all spans before a test run """
        self.recorder = async_tracer.recorder
        self.recorder.clear_spans()

        # New event loop for every test
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(None)

        # Restore default
        config['asyncio_task_context_propagation']['enabled'] = False

    def tearDown(self):
        """ Purge the queue """
        pass

    async def fetch(self, session, url, headers=None):
        try:
            async with session.get(url, headers=headers) as response:
                return response
        except aiohttp.web_exceptions.HTTPException:
            pass

    def test_ensure_future_with_context(self):
        async def run_later(msg="Hello"):
            # print("run_later: %s" % async_tracer.active_span.operation_name)
            async with aiohttp.ClientSession() as session:
                return await self.fetch(session, testenv["wsgi_server"] + "/")

        async def test():
            with async_tracer.start_active_span('test'):
                asyncio.ensure_future(run_later("Hello"))
            await asyncio.sleep(0.5)

        # Override default task context propagation
        config['asyncio_task_context_propagation']['enabled'] = True

        self.loop.run_until_complete(test())

        spans = self.recorder.queued_spans()
        self.assertEqual(3, len(spans))

        test_span = spans[0]
        wsgi_span = spans[1]
        aioclient_span = spans[2]

        self.assertEqual(test_span.t, wsgi_span.t)
        self.assertEqual(test_span.t, aioclient_span.t)

        self.assertEqual(test_span.p, None)
        self.assertEqual(wsgi_span.p, aioclient_span.s)
        self.assertEqual(aioclient_span.p, test_span.s)

    def test_ensure_future_without_context(self):
        async def run_later(msg="Hello"):
            # print("run_later: %s" % async_tracer.active_span.operation_name)
            async with aiohttp.ClientSession() as session:
                return await self.fetch(session, testenv["wsgi_server"] + "/")

        async def test():
            with async_tracer.start_active_span('test'):
                asyncio.ensure_future(run_later("Hello"))
            await asyncio.sleep(0.5)

        self.loop.run_until_complete(test())

        spans = self.recorder.queued_spans()

        self.assertEqual(2, len(spans))
        self.assertEqual("sdk", spans[0].n)
        self.assertEqual("wsgi", spans[1].n)

        # Without the context propagated, we should get two separate traces
        self.assertNotEqual(spans[0].t, spans[1].t)

    if hasattr(asyncio, "create_task"):
        def test_create_task_with_context(self):
            async def run_later(msg="Hello"):
                # print("run_later: %s" % async_tracer.active_span.operation_name)
                async with aiohttp.ClientSession() as session:
                    return await self.fetch(session, testenv["wsgi_server"] + "/")

            async def test():
                with async_tracer.start_active_span('test'):
                    asyncio.create_task(run_later("Hello"))
                await asyncio.sleep(0.5)

            # Override default task context propagation
            config['asyncio_task_context_propagation']['enabled'] = True

            self.loop.run_until_complete(test())

            spans = self.recorder.queued_spans()
            self.assertEqual(3, len(spans))

            test_span = spans[0]
            wsgi_span = spans[1]
            aioclient_span = spans[2]

            self.assertEqual(test_span.t, wsgi_span.t)
            self.assertEqual(test_span.t, aioclient_span.t)

            self.assertEqual(test_span.p, None)
            self.assertEqual(wsgi_span.p, aioclient_span.s)
            self.assertEqual(aioclient_span.p, test_span.s)

        def test_create_task_without_context(self):
            async def run_later(msg="Hello"):
                # print("run_later: %s" % async_tracer.active_span.operation_name)
                async with aiohttp.ClientSession() as session:
                    return await self.fetch(session, testenv["wsgi_server"] + "/")

            async def test():
                with async_tracer.start_active_span('test'):
                    asyncio.create_task(run_later("Hello"))
                await asyncio.sleep(0.5)

            self.loop.run_until_complete(test())

            spans = self.recorder.queued_spans()

            self.assertEqual(2, len(spans))
            self.assertEqual("sdk", spans[0].n)
            self.assertEqual("wsgi", spans[1].n)

            # Without the context propagated, we should get two separate traces
            self.assertNotEqual(spans[0].t, spans[1].t)



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2021

from instana.collector.helpers.process import ProcessHelper
from instana.log import logger


class FargateProcessHelper(ProcessHelper):
    """ Helper class to extend the generic process helper class with the corresponding fargate attributes """

    def collect_metrics(self, **kwargs):
        plugin_data = dict()
        try:
            plugin_data = super(FargateProcessHelper, self).collect_metrics(**kwargs)
            plugin_data["data"]["containerType"] = "docker"
            if self.collector.root_metadata is not None:
                plugin_data["data"]["container"] = self.collector.root_metadata.get("DockerId")

            if kwargs.get("with_snapshot"):
                if self.collector.task_metadata is not None:
                    plugin_data["data"]["com.instana.plugin.host.name"] = self.collector.task_metadata.get("TaskArn")
        except Exception:
            logger.debug("FargateProcessHelper.collect_metrics: ", exc_info=True)
        return [plugin_data]



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

"""
Instrumentation for FastAPI
https://fastapi.tiangolo.com/
"""
try:
    import fastapi
    import os
    import wrapt
    import signal
    from distutils.version import LooseVersion

    from ..log import logger
    from ..util.gunicorn import running_in_gunicorn
    from .asgi import InstanaASGIMiddleware
    from starlette.middleware import Middleware
    from fastapi import HTTPException
    from fastapi.exception_handlers import http_exception_handler

    from instana.singletons import async_tracer

    if hasattr(fastapi, '__version__') and \
        (LooseVersion(fastapi.__version__) >= LooseVersion('0.51.0')):

        async def instana_exception_handler(request, exc):
            """
            We capture FastAPI HTTPException, log the error and pass it on
            to the default exception handler.
            """
            try:
                span = async_tracer.active_span

                if span is not None:
                    if hasattr(exc, 'detail') and (500 <= exc.status_code <= 599):
                        span.set_tag('http.error', exc.detail)
                    span.set_tag('http.status_code', exc.status_code)
            except Exception:
                logger.debug("FastAPI instana_exception_handler: ", exc_info=True)

            return await http_exception_handler(request, exc)

        @wrapt.patch_function_wrapper('fastapi.applications', 'FastAPI.__init__')
        def init_with_instana(wrapped, instance, args, kwargs):
            middleware = kwargs.get('middleware')
            if middleware is None:
                kwargs['middleware'] = [Middleware(InstanaASGIMiddleware)]
            elif isinstance(middleware, list):
                middleware.append(Middleware(InstanaASGIMiddleware))

            exception_handlers = kwargs.get('exception_handlers')
            if exception_handlers is None:
                kwargs['exception_handlers'] = dict()

            if isinstance(kwargs['exception_handlers'], dict):
                kwargs['exception_handlers'][HTTPException] = instana_exception_handler

            return wrapped(*args, **kwargs)

        logger.debug("Instrumenting FastAPI")

        # Reload GUnicorn when we are instrumenting an already running application
        if "INSTANA_MAGIC" in os.environ and running_in_gunicorn():
            os.kill(os.getpid(), signal.SIGHUP)
    else:
        logger.debug("Instana supports FastAPI package versions 0.51.0 and newer.  Skipping.")

except ImportError:
    pass



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import inspect

import opentracing as ot

import instana.propagators.http_propagator as ihp
import instana.propagators.text_propagator as itp
import instana.propagators.binary_propagator as ibp
from instana.span_context import SpanContext
from instana.tracer import InstanaTracer


def test_http_basics():
    inspect.isclass(ihp.HTTPPropagator)

    inject_func = getattr(ihp.HTTPPropagator, "inject", None)
    assert inject_func
    assert callable(inject_func)

    extract_func = getattr(ihp.HTTPPropagator, "extract", None)
    assert extract_func
    assert callable(extract_func)


def test_http_inject_with_dict():
    ot.tracer = InstanaTracer()

    carrier = {}
    span = ot.tracer.start_span("nosetests")
    ot.tracer.inject(span.context, ot.Format.HTTP_HEADERS, carrier)

    assert 'X-INSTANA-T' in carrier
    assert carrier['X-INSTANA-T'] == span.context.trace_id
    assert 'X-INSTANA-S' in carrier
    assert carrier['X-INSTANA-S'] == span.context.span_id
    assert 'X-INSTANA-L' in carrier
    assert carrier['X-INSTANA-L'] == "1"


def test_http_inject_with_list():
    ot.tracer = InstanaTracer()

    carrier = []
    span = ot.tracer.start_span("nosetests")
    ot.tracer.inject(span.context, ot.Format.HTTP_HEADERS, carrier)

    assert ('X-INSTANA-T', span.context.trace_id) in carrier
    assert ('X-INSTANA-S', span.context.span_id) in carrier
    assert ('X-INSTANA-L', "1") in carrier


def test_http_basic_extract():
    ot.tracer = InstanaTracer()

    carrier = {'X-INSTANA-T': '1', 'X-INSTANA-S': '1', 'X-INSTANA-L': '1', 'X-INSTANA-SYNTHETIC': '1'}
    ctx = ot.tracer.extract(ot.Format.HTTP_HEADERS, carrier)

    assert isinstance(ctx, SpanContext)
    assert ctx.trace_id == '0000000000000001'
    assert ctx.span_id == '0000000000000001'
    assert ctx.synthetic


def test_http_extract_with_byte_keys():
    ot.tracer = InstanaTracer()

    carrier = {b'X-INSTANA-T': '1', b'X-INSTANA-S': '1', b'X-INSTANA-L': '1', b'X-INSTANA-SYNTHETIC': '1'}
    ctx = ot.tracer.extract(ot.Format.HTTP_HEADERS, carrier)

    assert isinstance(ctx, SpanContext)
    assert ctx.trace_id == '0000000000000001'
    assert ctx.span_id == '0000000000000001'
    assert ctx.synthetic


def test_http_extract_from_list_of_tuples():
    ot.tracer = InstanaTracer()

    carrier = [(b'user-agent', b'python-requests/2.23.0'), (b'accept-encoding', b'gzip, deflate'),
               (b'accept', b'*/*'), (b'connection', b'keep-alive'),
               (b'x-instana-t', b'1'), (b'x-instana-s', b'1'), (b'x-instana-l', b'1'), (b'X-INSTANA-SYNTHETIC', '1')]
    ctx = ot.tracer.extract(ot.Format.HTTP_HEADERS, carrier)

    assert isinstance(ctx, SpanContext)
    assert ctx.trace_id == '0000000000000001'
    assert ctx.span_id == '0000000000000001'
    assert ctx.synthetic


def test_http_mixed_case_extract():
    ot.tracer = InstanaTracer()

    carrier = {'x-insTana-T': '1', 'X-inSTANa-S': '1', 'X-INstana-l': '1'}
    ctx = ot.tracer.extract(ot.Format.HTTP_HEADERS, carrier)

    assert isinstance(ctx, SpanContext)
    assert ctx.trace_id == '0000000000000001'
    assert ctx.span_id == '0000000000000001'
    assert not ctx.synthetic


def test_http_extract_synthetic_only():
    ot.tracer = InstanaTracer()

    carrier = {'X-INSTANA-SYNTHETIC': '1'}
    ctx = ot.tracer.extract(ot.Format.HTTP_HEADERS, carrier)

    assert isinstance(ctx, SpanContext)
    assert ctx.trace_id is None
    assert ctx.span_id is None
    assert ctx.synthetic


def test_http_default_context_extract():
    ot.tracer = InstanaTracer()

    carrier = {}
    ctx = ot.tracer.extract(ot.Format.HTTP_HEADERS, carrier)

    assert isinstance(ctx, SpanContext)
    assert ctx.trace_id is None
    assert ctx.span_id is None
    assert ctx.synthetic is False

def test_http_128bit_headers():
    ot.tracer = InstanaTracer()

    carrier = {'X-INSTANA-T': '0000000000000000b0789916ff8f319f',
               'X-INSTANA-S': '0000000000000000b0789916ff8f319f', 'X-INSTANA-L': '1'}
    ctx = ot.tracer.extract(ot.Format.HTTP_HEADERS, carrier)

    assert isinstance(ctx, SpanContext)
    assert ctx.trace_id == 'b0789916ff8f319f'
    assert ctx.span_id == 'b0789916ff8f319f'


def test_text_basics():
    inspect.isclass(itp.TextPropagator)

    inject_func = getattr(itp.TextPropagator, "inject", None)
    assert inject_func
    assert callable(inject_func)

    extract_func = getattr(itp.TextPropagator, "extract", None)
    assert extract_func
    assert callable(extract_func)


def test_text_inject_with_dict():
    ot.tracer = InstanaTracer()

    carrier = {}
    span = ot.tracer.start_span("nosetests")
    ot.tracer.inject(span.context, ot.Format.TEXT_MAP, carrier)

    assert 'x-instana-t' in carrier
    assert carrier['x-instana-t'] == span.context.trace_id
    assert 'x-instana-s' in carrier
    assert carrier['x-instana-s'] == span.context.span_id
    assert 'x-instana-l' in carrier
    assert carrier['x-instana-l'] == "1"


def test_text_inject_with_list():
    ot.tracer = InstanaTracer()

    carrier = []
    span = ot.tracer.start_span("nosetests")
    ot.tracer.inject(span.context, ot.Format.TEXT_MAP, carrier)

    assert ('x-instana-t', span.context.trace_id) in carrier
    assert ('x-instana-s', span.context.span_id) in carrier
    assert ('x-instana-l', "1") in carrier


def test_text_basic_extract():
    ot.tracer = InstanaTracer()

    carrier = {'x-instana-t': '1', 'x-instana-s': '1', 'x-instana-l': '1'}
    ctx = ot.tracer.extract(ot.Format.TEXT_MAP, carrier)

    assert isinstance(ctx, SpanContext)
    assert ctx.trace_id == '0000000000000001'
    assert ctx.span_id == '0000000000000001'


def test_text_mixed_case_extract():
    ot.tracer = InstanaTracer()

    carrier = {'x-insTana-T': '1', 'X-inSTANa-S': '1', 'X-INstana-l': '1'}
    ctx = ot.tracer.extract(ot.Format.TEXT_MAP, carrier)

    assert isinstance(ctx, SpanContext)
    assert ctx.trace_id == '0000000000000001'
    assert ctx.span_id == '0000000000000001'


def test_text_default_context_extract():
    ot.tracer = InstanaTracer()

    carrier = {}
    ctx = ot.tracer.extract(ot.Format.TEXT_MAP, carrier)

    assert isinstance(ctx, SpanContext)
    assert ctx.trace_id is None
    assert ctx.span_id is None
    assert ctx.synthetic is False


def test_text_128bit_headers():
    ot.tracer = InstanaTracer()

    carrier = {'x-instana-t': '0000000000000000b0789916ff8f319f',
               'x-instana-s': ' 0000000000000000b0789916ff8f319f', 'X-INSTANA-L': '1'}
    ctx = ot.tracer.extract(ot.Format.TEXT_MAP, carrier)

    assert isinstance(ctx, SpanContext)
    assert('b0789916ff8f319f' == ctx.span_id)
    assert ctx.trace_id == 'b0789916ff8f319f'
    assert ctx.span_id == 'b0789916ff8f319f'

def test_binary_basics():
    inspect.isclass(ibp.BinaryPropagator)

    inject_func = getattr(ibp.BinaryPropagator, "inject", None)
    assert inject_func
    assert callable(inject_func)

    extract_func = getattr(ibp.BinaryPropagator, "extract", None)
    assert extract_func
    assert callable(extract_func)


def test_binary_inject_with_dict():
    ot.tracer = InstanaTracer()

    carrier = {}
    span = ot.tracer.start_span("nosetests")
    ot.tracer.inject(span.context, ot.Format.BINARY, carrier)

    assert b'x-instana-t' in carrier
    assert carrier[b'x-instana-t'] == str.encode(span.context.trace_id)
    assert b'x-instana-s' in carrier
    assert carrier[b'x-instana-s'] == str.encode(span.context.span_id)
    assert b'x-instana-l' in carrier
    assert carrier[b'x-instana-l'] == b'1'


def test_binary_inject_with_list():
    ot.tracer = InstanaTracer()

    carrier = []
    span = ot.tracer.start_span("nosetests")
    ot.tracer.inject(span.context, ot.Format.BINARY, carrier)

    assert (b'x-instana-t', str.encode(span.context.trace_id)) in carrier
    assert (b'x-instana-s', str.encode(span.context.span_id)) in carrier
    assert (b'x-instana-l', b'1') in carrier


def test_binary_basic_extract():
    ot.tracer = InstanaTracer()

    carrier = {b'X-INSTANA-T': b'1', b'X-INSTANA-S': b'1', b'X-INSTANA-L': b'1', b'X-INSTANA-SYNTHETIC': b'1'}
    ctx = ot.tracer.extract(ot.Format.BINARY, carrier)

    assert isinstance(ctx, SpanContext)
    assert ctx.trace_id == '0000000000000001'
    assert ctx.span_id == '0000000000000001'
    assert ctx.synthetic


def test_binary_mixed_case_extract():
    ot.tracer = InstanaTracer()

    carrier = {'x-insTana-T': '1', 'X-inSTANa-S': '1', 'X-INstana-l': '1', b'X-inStaNa-SYNtheTIC': b'1'}
    ctx = ot.tracer.extract(ot.Format.BINARY, carrier)

    assert isinstance(ctx, SpanContext)
    assert ctx.trace_id == '0000000000000001'
    assert ctx.span_id == '0000000000000001'
    assert ctx.synthetic


def test_binary_default_context_extract():
    ot.tracer = InstanaTracer()

    carrier = {}
    ctx = ot.tracer.extract(ot.Format.BINARY, carrier)

    assert isinstance(ctx, SpanContext)
    assert ctx.trace_id is None
    assert ctx.span_id is None
    assert ctx.synthetic is False


def test_binary_128bit_headers():
    ot.tracer = InstanaTracer()

    carrier = {'X-INSTANA-T': '0000000000000000b0789916ff8f319f',
               'X-INSTANA-S': ' 0000000000000000b0789916ff8f319f', 'X-INSTANA-L': '1'}
    ctx = ot.tracer.extract(ot.Format.BINARY, carrier)

    assert isinstance(ctx, SpanContext)
    assert ctx.trace_id == 'b0789916ff8f319f'
    assert ctx.span_id == 'b0789916ff8f319f'



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2019


class SpanContext():
    def __init__(
            self,
            trace_id=None,
            span_id=None,
            baggage=None,
            sampled=True,
            level=1,
            synthetic=False
    ):

        self.level = level
        self.trace_id = trace_id
        self.span_id = span_id
        self.sampled = sampled
        self.synthetic = synthetic
        self._baggage = baggage or {}

        self.trace_parent = None  # true/false flag
        self.instana_ancestor = None
        self.long_trace_id = None
        self.correlation_type = None
        self.correlation_id = None
        self.traceparent = None  # temporary storage of the validated traceparent header of the incoming request
        self.tracestate = None  # temporary storage of the tracestate header

    @property
    def traceparent(self):
        return self._traceparent

    @traceparent.setter
    def traceparent(self, value):
        self._traceparent = value

    @property
    def tracestate(self):
        return self._tracestate

    @tracestate.setter
    def tracestate(self, value):
        self._tracestate = value

    @property
    def trace_parent(self):
        return self._trace_parent

    @trace_parent.setter
    def trace_parent(self, value):
        self._trace_parent = value

    @property
    def instana_ancestor(self):
        return self._instana_ancestor

    @instana_ancestor.setter
    def instana_ancestor(self, value):
        self._instana_ancestor = value

    @property
    def long_trace_id(self):
        return self._long_trace_id

    @long_trace_id.setter
    def long_trace_id(self, value):
        self._long_trace_id = value

    @property
    def correlation_type(self):
        return self._correlation_type

    @correlation_type.setter
    def correlation_type(self, value):
        self._correlation_type = value

    @property
    def correlation_id(self):
        return self._correlation_id

    @correlation_id.setter
    def correlation_id(self, value):
        self._correlation_id = value

    @property
    def baggage(self):
        return self._baggage

    def with_baggage_item(self, key, value):
        new_baggage = self._baggage.copy()
        new_baggage[key] = value
        return SpanContext(
            trace_id=self.trace_id,
            span_id=self.span_id,
            sampled=self.sampled,
            baggage=new_baggage)


# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

from __future__ import absolute_import

import sys
import logging
import unittest
from ..helpers import testenv
from unittest import SkipTest
from instana.singletons import tracer


if sys.version_info[0] > 2:
    import MySQLdb
else:
    raise SkipTest("mysqlclient supported on Python 3 only")

logger = logging.getLogger(__name__)

create_table_query = 'CREATE TABLE IF NOT EXISTS users(id serial primary key, \
                      name varchar(40) NOT NULL, email varchar(40) NOT NULL)'

create_proc_query = """
CREATE PROCEDURE test_proc(IN t VARCHAR(255))
BEGIN
    SELECT name FROM users WHERE name = t;
END
"""

db = MySQLdb.connect(host=testenv['mysql_host'], port=testenv['mysql_port'],
                     user=testenv['mysql_user'], passwd=testenv['mysql_pw'],
                     db=testenv['mysql_db'])

cursor = db.cursor()
cursor.execute(create_table_query)

while cursor.nextset() is not None:
    pass

cursor.execute('DROP PROCEDURE IF EXISTS test_proc')

while cursor.nextset() is not None:
    pass

cursor.execute(create_proc_query)

while cursor.nextset() is not None:
    pass

cursor.close()
db.close()


class TestMySQLPython(unittest.TestCase):
    def setUp(self):
        self.db = MySQLdb.connect(host=testenv['mysql_host'], port=testenv['mysql_port'],
                                  user=testenv['mysql_user'], passwd=testenv['mysql_pw'],
                                  db=testenv['mysql_db'])
        self.cursor = self.db.cursor()
        self.recorder = tracer.recorder
        self.recorder.clear_spans()
        tracer.cur_ctx = None

    def tearDown(self):
        """ Do nothing for now """
        return None

    def test_vanilla_query(self):
        self.cursor.execute("""SELECT * from users""")
        result = self.cursor.fetchone()
        self.assertEqual(3, len(result))

        spans = self.recorder.queued_spans()
        self.assertEqual(0, len(spans))

    def test_basic_query(self):
        result = None
        with tracer.start_active_span('test'):
            result = self.cursor.execute("""SELECT * from users""")
            self.cursor.fetchone()

        assert(result >= 0)

        spans = self.recorder.queued_spans()
        self.assertEqual(2, len(spans))

        db_span = spans[0]
        test_span = spans[1]

        self.assertEqual("test", test_span.data["sdk"]["name"])
        self.assertEqual(test_span.t, db_span.t)
        self.assertEqual(db_span.p, test_span.s)

        self.assertEqual(None, db_span.ec)

        self.assertEqual(db_span.n, "mysql")
        self.assertEqual(db_span.data["mysql"]["db"], testenv['mysql_db'])
        self.assertEqual(db_span.data["mysql"]["user"], testenv['mysql_user'])
        self.assertEqual(db_span.data["mysql"]["stmt"], 'SELECT * from users')
        self.assertEqual(db_span.data["mysql"]["host"], testenv['mysql_host'])
        self.assertEqual(db_span.data["mysql"]["port"], testenv['mysql_port'])

    def test_basic_insert(self):
        result = None
        with tracer.start_active_span('test'):
            result = self.cursor.execute(
                        """INSERT INTO users(name, email) VALUES(%s, %s)""",
                        ('beaker', 'beaker@muppets.com'))

        self.assertEqual(1, result)

        spans = self.recorder.queued_spans()
        self.assertEqual(2, len(spans))

        db_span = spans[0]
        test_span = spans[1]

        self.assertEqual("test", test_span.data["sdk"]["name"])
        self.assertEqual(test_span.t, db_span.t)
        self.assertEqual(db_span.p, test_span.s)

        self.assertEqual(None, db_span.ec)

        self.assertEqual(db_span.n, "mysql")
        self.assertEqual(db_span.data["mysql"]["db"], testenv['mysql_db'])
        self.assertEqual(db_span.data["mysql"]["user"], testenv['mysql_user'])
        self.assertEqual(db_span.data["mysql"]["stmt"], 'INSERT INTO users(name, email) VALUES(%s, %s)')
        self.assertEqual(db_span.data["mysql"]["host"], testenv['mysql_host'])
        self.assertEqual(db_span.data["mysql"]["port"], testenv['mysql_port'])

    def test_executemany(self):
        result = None
        with tracer.start_active_span('test'):
            result = self.cursor.executemany("INSERT INTO users(name, email) VALUES(%s, %s)",
                                             [('beaker', 'beaker@muppets.com'), ('beaker', 'beaker@muppets.com')])
            self.db.commit()

        self.assertEqual(2, result)

        spans = self.recorder.queued_spans()
        self.assertEqual(2, len(spans))

        db_span = spans[0]
        test_span = spans[1]

        self.assertEqual("test", test_span.data["sdk"]["name"])
        self.assertEqual(test_span.t, db_span.t)
        self.assertEqual(db_span.p, test_span.s)

        self.assertEqual(None, db_span.ec)

        self.assertEqual(db_span.n, "mysql")
        self.assertEqual(db_span.data["mysql"]["db"], testenv['mysql_db'])
        self.assertEqual(db_span.data["mysql"]["user"], testenv['mysql_user'])
        self.assertEqual(db_span.data["mysql"]["stmt"], 'INSERT INTO users(name, email) VALUES(%s, %s)')
        self.assertEqual(db_span.data["mysql"]["host"], testenv['mysql_host'])
        self.assertEqual(db_span.data["mysql"]["port"], testenv['mysql_port'])

    def test_call_proc(self):
        result = None
        with tracer.start_active_span('test'):
            result = self.cursor.callproc('test_proc', ('beaker',))

        assert(result)

        spans = self.recorder.queued_spans()
        self.assertEqual(2, len(spans))

        db_span = spans[0]
        test_span = spans[1]

        self.assertEqual("test", test_span.data["sdk"]["name"])
        self.assertEqual(test_span.t, db_span.t)
        self.assertEqual(db_span.p, test_span.s)

        self.assertEqual(None, db_span.ec)

        self.assertEqual(db_span.n, "mysql")
        self.assertEqual(db_span.data["mysql"]["db"], testenv['mysql_db'])
        self.assertEqual(db_span.data["mysql"]["user"], testenv['mysql_user'])
        self.assertEqual(db_span.data["mysql"]["stmt"], 'test_proc')
        self.assertEqual(db_span.data["mysql"]["host"], testenv['mysql_host'])
        self.assertEqual(db_span.data["mysql"]["port"], testenv['mysql_port'])

    def test_error_capture(self):
        result = None
        span = None
        try:
            with tracer.start_active_span('test'):
                result = self.cursor.execute("""SELECT * from blah""")
                self.cursor.fetchone()
        except Exception:
            pass
        finally:
            if span:
                span.finish()

        assert(result is None)

        spans = self.recorder.queued_spans()
        self.assertEqual(2, len(spans))

        db_span = spans[0]
        test_span = spans[1]

        self.assertEqual("test", test_span.data["sdk"]["name"])
        self.assertEqual(test_span.t, db_span.t)
        self.assertEqual(db_span.p, test_span.s)

        self.assertEqual(1, db_span.ec)
        self.assertEqual(db_span.data["mysql"]["error"], '(1146, "Table \'%s.blah\' doesn\'t exist")' % testenv['mysql_db'])

        self.assertEqual(db_span.n, "mysql")
        self.assertEqual(db_span.data["mysql"]["db"], testenv['mysql_db'])
        self.assertEqual(db_span.data["mysql"]["user"], testenv['mysql_user'])
        self.assertEqual(db_span.data["mysql"]["stmt"], 'SELECT * from blah')
        self.assertEqual(db_span.data["mysql"]["host"], testenv['mysql_host'])
        self.assertEqual(db_span.data["mysql"]["port"], testenv['mysql_port'])



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

from __future__ import absolute_import

import time
import urllib3
import unittest

import tests.apps.flask_app
from ..helpers import testenv
from instana.singletons import agent, tracer


class TestWSGI(unittest.TestCase):
    def setUp(self):
        """ Clear all spans before a test run """
        self.http = urllib3.PoolManager()
        self.recorder = tracer.recorder
        self.recorder.clear_spans()
        time.sleep(0.1)

    def tearDown(self):
        """ Do nothing for now """
        return None

    def test_vanilla_requests(self):
        response = self.http.request('GET', testenv["wsgi_server"] + '/')
        spans = self.recorder.queued_spans()

        self.assertEqual(1, len(spans))
        self.assertIsNone(tracer.active_span)
        self.assertEqual(response.status, 200)

    def test_get_request(self):
        with tracer.start_active_span('test'):
            response = self.http.request('GET', testenv["wsgi_server"] + '/')

        spans = self.recorder.queued_spans()

        self.assertEqual(3, len(spans))
        self.assertIsNone(tracer.active_span)

        wsgi_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        assert response
        self.assertEqual(200, response.status)

        assert 'X-INSTANA-T' in response.headers
        assert(int(response.headers['X-INSTANA-T'], 16))
        self.assertEqual(response.headers['X-INSTANA-T'], wsgi_span.t)

        assert 'X-INSTANA-S' in response.headers
        assert(int(response.headers['X-INSTANA-S'], 16))
        self.assertEqual(response.headers['X-INSTANA-S'], wsgi_span.s)

        assert 'X-INSTANA-L' in response.headers
        self.assertEqual(response.headers['X-INSTANA-L'], '1')

        assert 'Server-Timing' in response.headers
        server_timing_value = "intid;desc=%s" % wsgi_span.t
        self.assertEqual(response.headers['Server-Timing'], server_timing_value)

        # Same traceId
        self.assertEqual(test_span.t, urllib3_span.t)
        self.assertEqual(urllib3_span.t, wsgi_span.t)

        # Parent relationships
        self.assertEqual(urllib3_span.p, test_span.s)
        self.assertEqual(wsgi_span.p, urllib3_span.s)

        self.assertIsNone(wsgi_span.sy)
        self.assertIsNone(urllib3_span.sy)
        self.assertIsNone(test_span.sy)

        # Error logging
        self.assertIsNone(test_span.ec)
        self.assertIsNone(urllib3_span.ec)
        self.assertIsNone(wsgi_span.ec)

        # wsgi
        self.assertEqual("wsgi", wsgi_span.n)
        self.assertEqual('127.0.0.1:' + str(testenv['wsgi_port']), wsgi_span.data["http"]["host"])
        self.assertEqual('/', wsgi_span.data["http"]["url"])
        self.assertEqual('GET', wsgi_span.data["http"]["method"])
        self.assertEqual(200, wsgi_span.data["http"]["status"])
        self.assertIsNone(wsgi_span.data["http"]["error"])
        self.assertIsNone(wsgi_span.stack)

    def test_synthetic_request(self):
        headers = {
            'X-INSTANA-SYNTHETIC': '1'
        }
        with tracer.start_active_span('test'):
            response = self.http.request('GET', testenv["wsgi_server"] + '/', headers=headers)

        spans = self.recorder.queued_spans()

        self.assertEqual(3, len(spans))
        self.assertIsNone(tracer.active_span)

        wsgi_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        self.assertTrue(wsgi_span.sy)
        self.assertIsNone(urllib3_span.sy)
        self.assertIsNone(test_span.sy)

    def test_complex_request(self):
        with tracer.start_active_span('test'):
            response = self.http.request('GET', testenv["wsgi_server"] + '/complex')

        spans = self.recorder.queued_spans()
        self.assertEqual(5, len(spans))
        self.assertIsNone(tracer.active_span)

        spacedust_span = spans[0]
        asteroid_span = spans[1]
        wsgi_span = spans[2]
        urllib3_span = spans[3]
        test_span = spans[4]

        assert response
        self.assertEqual(200, response.status)

        assert 'X-INSTANA-T' in response.headers
        assert(int(response.headers['X-INSTANA-T'], 16))
        self.assertEqual(response.headers['X-INSTANA-T'], wsgi_span.t)

        assert 'X-INSTANA-S' in response.headers
        assert(int(response.headers['X-INSTANA-S'], 16))
        self.assertEqual(response.headers['X-INSTANA-S'], wsgi_span.s)

        assert 'X-INSTANA-L' in response.headers
        self.assertEqual(response.headers['X-INSTANA-L'], '1')

        assert 'Server-Timing' in response.headers
        server_timing_value = "intid;desc=%s" % wsgi_span.t
        self.assertEqual(response.headers['Server-Timing'], server_timing_value)

        # Same traceId
        trace_id = test_span.t
        self.assertEqual(trace_id, urllib3_span.t)
        self.assertEqual(trace_id, wsgi_span.t)
        self.assertEqual(trace_id, asteroid_span.t)
        self.assertEqual(trace_id, spacedust_span.t)

        # Parent relationships
        self.assertEqual(urllib3_span.p, test_span.s)
        self.assertEqual(wsgi_span.p, urllib3_span.s)
        self.assertEqual(asteroid_span.p, wsgi_span.s)
        self.assertEqual(spacedust_span.p, asteroid_span.s)

        # Error logging
        self.assertIsNone(test_span.ec)
        self.assertIsNone(urllib3_span.ec)
        self.assertIsNone(wsgi_span.ec)
        self.assertIsNone(asteroid_span.ec)
        self.assertIsNone(spacedust_span.ec)

        # wsgi
        self.assertEqual("wsgi", wsgi_span.n)
        self.assertEqual('127.0.0.1:' + str(testenv['wsgi_port']), wsgi_span.data["http"]["host"])
        self.assertEqual('/complex', wsgi_span.data["http"]["url"])
        self.assertEqual('GET', wsgi_span.data["http"]["method"])
        self.assertEqual(200, wsgi_span.data["http"]["status"])
        self.assertIsNone(wsgi_span.data["http"]["error"])
        self.assertIsNone(wsgi_span.stack)

    def test_custom_header_capture(self):
        # Hack together a manual custom headers list
        agent.options.extra_http_headers = [u'X-Capture-This', u'X-Capture-That']

        request_headers = {}
        request_headers['X-Capture-This'] = 'this'
        request_headers['X-Capture-That'] = 'that'

        with tracer.start_active_span('test'):
            response = self.http.request('GET', testenv["wsgi_server"] + '/', headers=request_headers)

        spans = self.recorder.queued_spans()

        self.assertEqual(3, len(spans))
        self.assertIsNone(tracer.active_span)

        wsgi_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        assert response
        self.assertEqual(200, response.status)

        assert 'X-INSTANA-T' in response.headers
        assert(int(response.headers['X-INSTANA-T'], 16))
        self.assertEqual(response.headers['X-INSTANA-T'], wsgi_span.t)

        assert 'X-INSTANA-S' in response.headers
        assert(int(response.headers['X-INSTANA-S'], 16))
        self.assertEqual(response.headers['X-INSTANA-S'], wsgi_span.s)

        assert 'X-INSTANA-L' in response.headers
        self.assertEqual(response.headers['X-INSTANA-L'], '1')

        assert 'Server-Timing' in response.headers
        server_timing_value = "intid;desc=%s" % wsgi_span.t
        self.assertEqual(response.headers['Server-Timing'], server_timing_value)

        # Same traceId
        self.assertEqual(test_span.t, urllib3_span.t)
        self.assertEqual(urllib3_span.t, wsgi_span.t)

        # Parent relationships
        self.assertEqual(urllib3_span.p, test_span.s)
        self.assertEqual(wsgi_span.p, urllib3_span.s)

        # Error logging
        self.assertIsNone(test_span.ec)
        self.assertIsNone(urllib3_span.ec)
        self.assertIsNone(wsgi_span.ec)

        # wsgi
        self.assertEqual("wsgi", wsgi_span.n)
        self.assertEqual('127.0.0.1:' + str(testenv['wsgi_port']), wsgi_span.data["http"]["host"])
        self.assertEqual('/', wsgi_span.data["http"]["url"])
        self.assertEqual('GET', wsgi_span.data["http"]["method"])
        self.assertEqual(200, wsgi_span.data["http"]["status"])
        self.assertIsNone(wsgi_span.data["http"]["error"])
        self.assertIsNone(wsgi_span.stack)

        assert "X-Capture-This" in wsgi_span.data["http"]["header"]
        self.assertEqual("this", wsgi_span.data["http"]["header"]["X-Capture-This"])
        assert "X-Capture-That" in wsgi_span.data["http"]["header"]
        self.assertEqual("that", wsgi_span.data["http"]["header"]["X-Capture-That"])

    def test_secret_scrubbing(self):
        with tracer.start_active_span('test'):
            response = self.http.request('GET', testenv["wsgi_server"] + '/?secret=shhh')

        spans = self.recorder.queued_spans()

        self.assertEqual(3, len(spans))
        self.assertIsNone(tracer.active_span)

        wsgi_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        assert response
        self.assertEqual(200, response.status)

        assert 'X-INSTANA-T' in response.headers
        assert(int(response.headers['X-INSTANA-T'], 16))
        self.assertEqual(response.headers['X-INSTANA-T'], wsgi_span.t)

        assert 'X-INSTANA-S' in response.headers
        assert(int(response.headers['X-INSTANA-S'], 16))
        self.assertEqual(response.headers['X-INSTANA-S'], wsgi_span.s)

        assert 'X-INSTANA-L' in response.headers
        self.assertEqual(response.headers['X-INSTANA-L'], '1')

        assert 'Server-Timing' in response.headers
        server_timing_value = "intid;desc=%s" % wsgi_span.t
        self.assertEqual(response.headers['Server-Timing'], server_timing_value)

        # Same traceId
        self.assertEqual(test_span.t, urllib3_span.t)
        self.assertEqual(urllib3_span.t, wsgi_span.t)

        # Parent relationships
        self.assertEqual(urllib3_span.p, test_span.s)
        self.assertEqual(wsgi_span.p, urllib3_span.s)

        # Error logging
        self.assertIsNone(test_span.ec)
        self.assertIsNone(urllib3_span.ec)
        self.assertIsNone(wsgi_span.ec)

        # wsgi
        self.assertEqual("wsgi", wsgi_span.n)
        self.assertEqual('127.0.0.1:' + str(testenv['wsgi_port']), wsgi_span.data["http"]["host"])
        self.assertEqual('/', wsgi_span.data["http"]["url"])
        self.assertEqual('secret=<redacted>', wsgi_span.data["http"]["params"])
        self.assertEqual('GET', wsgi_span.data["http"]["method"])
        self.assertEqual(200, wsgi_span.data["http"]["status"])
        self.assertIsNone(wsgi_span.data["http"]["error"])
        self.assertIsNone(wsgi_span.stack)

    def test_with_incoming_context(self):
        request_headers = dict()
        request_headers['X-INSTANA-T'] = '0000000000000001'
        request_headers['X-INSTANA-S'] = '0000000000000001'

        response = self.http.request('GET', testenv["wsgi_server"] + '/', headers=request_headers)

        assert response
        self.assertEqual(200, response.status)

        spans = self.recorder.queued_spans()
        self.assertEqual(1, len(spans))

        wsgi_span = spans[0]

        self.assertEqual(wsgi_span.t, '0000000000000001')
        self.assertEqual(wsgi_span.p, '0000000000000001')

        assert 'X-INSTANA-T' in response.headers
        assert(int(response.headers['X-INSTANA-T'], 16))
        self.assertEqual(response.headers['X-INSTANA-T'], wsgi_span.t)

        assert 'X-INSTANA-S' in response.headers
        assert(int(response.headers['X-INSTANA-S'], 16))
        self.assertEqual(response.headers['X-INSTANA-S'], wsgi_span.s)

        assert 'X-INSTANA-L' in response.headers
        self.assertEqual(response.headers['X-INSTANA-L'], '1')

        assert 'Server-Timing' in response.headers
        server_timing_value = "intid;desc=%s" % wsgi_span.t
        self.assertEqual(response.headers['Server-Timing'], server_timing_value)

    def test_with_incoming_mixed_case_context(self):
        request_headers = dict()
        request_headers['X-InSTANa-T'] = '0000000000000001'
        request_headers['X-instana-S'] = '0000000000000001'

        response = self.http.request('GET', testenv["wsgi_server"] + '/', headers=request_headers)

        assert response
        self.assertEqual(200, response.status)

        spans = self.recorder.queued_spans()
        self.assertEqual(1, len(spans))

        wsgi_span = spans[0]

        self.assertEqual(wsgi_span.t, '0000000000000001')
        self.assertEqual(wsgi_span.p, '0000000000000001')

        assert 'X-INSTANA-T' in response.headers
        assert(int(response.headers['X-INSTANA-T'], 16))
        self.assertEqual(response.headers['X-INSTANA-T'], wsgi_span.t)

        assert 'X-INSTANA-S' in response.headers
        assert(int(response.headers['X-INSTANA-S'], 16))
        self.assertEqual(response.headers['X-INSTANA-S'], wsgi_span.s)

        assert 'X-INSTANA-L' in response.headers
        self.assertEqual(response.headers['X-INSTANA-L'], '1')

        assert 'Server-Timing' in response.headers
        server_timing_value = "intid;desc=%s" % wsgi_span.t
        self.assertEqual(response.headers['Server-Timing'], server_timing_value)

    def test_response_headers(self):
        with tracer.start_active_span('test'):
            response = self.http.request('GET', testenv["wsgi_server"] + '/')

        spans = self.recorder.queued_spans()

        self.assertEqual(3, len(spans))
        self.assertIsNone(tracer.active_span)

        wsgi_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        assert response
        self.assertEqual(200, response.status)

        assert 'X-INSTANA-T' in response.headers
        assert(int(response.headers['X-INSTANA-T'], 16))
        self.assertEqual(response.headers['X-INSTANA-T'], wsgi_span.t)

        assert 'X-INSTANA-S' in response.headers
        assert(int(response.headers['X-INSTANA-S'], 16))
        self.assertEqual(response.headers['X-INSTANA-S'], wsgi_span.s)

        assert 'X-INSTANA-L' in response.headers
        self.assertEqual(response.headers['X-INSTANA-L'], '1')

        assert 'Server-Timing' in response.headers
        server_timing_value = "intid;desc=%s" % wsgi_span.t
        self.assertEqual(response.headers['Server-Timing'], server_timing_value)



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

from __future__ import absolute_import

import unittest
import urllib3

import tests.apps.pyramid_app
from ..helpers import testenv
from instana.singletons import tracer


class TestPyramid(unittest.TestCase):
    def setUp(self):
        """ Clear all spans before a test run """
        self.http = urllib3.PoolManager()
        self.recorder = tracer.recorder
        self.recorder.clear_spans()

    def tearDown(self):
        """ Do nothing for now """
        return None

    def test_vanilla_requests(self):
        r = self.http.request('GET', testenv["pyramid_server"] + '/')
        self.assertEqual(r.status, 200)

        spans = self.recorder.queued_spans()
        self.assertEqual(1, len(spans))

    def test_get_request(self):
        with tracer.start_active_span('test'):
            response = self.http.request('GET', testenv["pyramid_server"] + '/')

        spans = self.recorder.queued_spans()
        self.assertEqual(3, len(spans))

        pyramid_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        assert response
        self.assertEqual(200, response.status)

        assert('X-INSTANA-T' in response.headers)
        assert(int(response.headers['X-INSTANA-T'], 16))
        self.assertEqual(response.headers['X-INSTANA-T'], pyramid_span.t)

        assert('X-INSTANA-S' in response.headers)
        assert(int(response.headers['X-INSTANA-S'], 16))
        self.assertEqual(response.headers['X-INSTANA-S'], pyramid_span.s)

        assert('X-INSTANA-L' in response.headers)
        self.assertEqual(response.headers['X-INSTANA-L'], '1')

        assert('Server-Timing' in response.headers)
        server_timing_value = "intid;desc=%s" % pyramid_span.t
        self.assertEqual(response.headers['Server-Timing'], server_timing_value)

        self.assertIsNone(tracer.active_span)

        # Same traceId
        self.assertEqual(test_span.t, urllib3_span.t)
        self.assertEqual(urllib3_span.t, pyramid_span.t)

        # Parent relationships
        self.assertEqual(urllib3_span.p, test_span.s)
        self.assertEqual(pyramid_span.p, urllib3_span.s)

        # Synthetic
        self.assertIsNone(pyramid_span.sy)
        self.assertIsNone(urllib3_span.sy)
        self.assertIsNone(test_span.sy)

        # Error logging
        self.assertIsNone(test_span.ec)
        self.assertIsNone(urllib3_span.ec)
        self.assertIsNone(pyramid_span.ec)

        # HTTP SDK span
        self.assertEqual("sdk", pyramid_span.n)

        assert(pyramid_span.data["sdk"])
        self.assertEqual('http', pyramid_span.data["sdk"]["name"])
        self.assertEqual('entry', pyramid_span.data["sdk"]["type"])

        sdk_data = pyramid_span.data["sdk"]["custom"]
        self.assertEqual('127.0.0.1:' + str(testenv['pyramid_port']), sdk_data["tags"]["http.host"])
        self.assertEqual('/', sdk_data["tags"]["http.url"])
        self.assertEqual('GET', sdk_data["tags"]["http.method"])
        self.assertEqual(200, sdk_data["tags"]["http.status"])
        self.assertNotIn("message", sdk_data["tags"])
        self.assertNotIn("http.path_tpl", sdk_data["tags"])

        # urllib3
        self.assertEqual("test", test_span.data["sdk"]["name"])
        self.assertEqual("urllib3", urllib3_span.n)
        self.assertEqual(200, urllib3_span.data["http"]["status"])
        self.assertEqual(testenv["pyramid_server"] + '/', urllib3_span.data["http"]["url"])
        self.assertEqual("GET", urllib3_span.data["http"]["method"])
        self.assertIsNotNone(urllib3_span.stack)
        self.assertTrue(type(urllib3_span.stack) is list)
        self.assertTrue(len(urllib3_span.stack) > 1)

    def test_synthetic_request(self):
        headers = {
            'X-INSTANA-SYNTHETIC': '1'
        }

        with tracer.start_active_span('test'):
            response = self.http.request('GET', testenv["pyramid_server"] + '/', headers=headers)

        spans = self.recorder.queued_spans()
        self.assertEqual(3, len(spans))

        pyramid_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        assert response
        self.assertEqual(200, response.status)

        self.assertTrue(pyramid_span.sy)
        self.assertIsNone(urllib3_span.sy)
        self.assertIsNone(test_span.sy)

    def test_500(self):
        with tracer.start_active_span('test'):
            response = self.http.request('GET', testenv["pyramid_server"] + '/500')

        spans = self.recorder.queued_spans()

        self.assertEqual(3, len(spans))

        pyramid_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        assert response
        self.assertEqual(500, response.status)

        assert('X-INSTANA-T' in response.headers)
        assert(int(response.headers['X-INSTANA-T'], 16))
        self.assertEqual(response.headers['X-INSTANA-T'], pyramid_span.t)

        assert('X-INSTANA-S' in response.headers)
        assert(int(response.headers['X-INSTANA-S'], 16))
        self.assertEqual(response.headers['X-INSTANA-S'], pyramid_span.s)

        assert('X-INSTANA-L' in response.headers)
        self.assertEqual(response.headers['X-INSTANA-L'], '1')

        assert('Server-Timing' in response.headers)
        server_timing_value = "intid;desc=%s" % pyramid_span.t
        self.assertEqual(response.headers['Server-Timing'], server_timing_value)

        self.assertIsNone(tracer.active_span)

        # Same traceId
        self.assertEqual(test_span.t, urllib3_span.t)
        self.assertEqual(test_span.t, pyramid_span.t)

        # Parent relationships
        self.assertEqual(urllib3_span.p, test_span.s)
        self.assertEqual(pyramid_span.p, urllib3_span.s)

        # Error logging
        self.assertIsNone(test_span.ec)
        self.assertEqual(1, urllib3_span.ec)
        self.assertEqual(1, pyramid_span.ec)

        # wsgi
        self.assertEqual("sdk", pyramid_span.n)
        self.assertEqual('http', pyramid_span.data["sdk"]["name"])
        self.assertEqual('entry', pyramid_span.data["sdk"]["type"])

        sdk_data = pyramid_span.data["sdk"]["custom"]
        self.assertEqual('127.0.0.1:' + str(testenv['pyramid_port']), sdk_data["tags"]["http.host"])
        self.assertEqual('/500', sdk_data["tags"]["http.url"])
        self.assertEqual('GET', sdk_data["tags"]["http.method"])
        self.assertEqual(500, sdk_data["tags"]["http.status"])
        self.assertEqual("internal error", sdk_data["tags"]["message"])
        self.assertNotIn("http.path_tpl", sdk_data["tags"])

        # urllib3
        self.assertEqual("test", test_span.data["sdk"]["name"])
        self.assertEqual("urllib3", urllib3_span.n)
        self.assertEqual(500, urllib3_span.data["http"]["status"])
        self.assertEqual(testenv["pyramid_server"] + '/500', urllib3_span.data["http"]["url"])
        self.assertEqual("GET", urllib3_span.data["http"]["method"])
        self.assertIsNotNone(urllib3_span.stack)
        self.assertTrue(type(urllib3_span.stack) is list)
        self.assertTrue(len(urllib3_span.stack) > 1)

    def test_exception(self):
        with tracer.start_active_span('test'):
            response = self.http.request('GET', testenv["pyramid_server"] + '/exception')

        spans = self.recorder.queued_spans()

        self.assertEqual(3, len(spans))

        pyramid_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        assert response
        self.assertEqual(500, response.status)

        self.assertIsNone(tracer.active_span)

        # Same traceId
        self.assertEqual(test_span.t, urllib3_span.t)
        self.assertEqual(test_span.t, pyramid_span.t)

        # Parent relationships
        self.assertEqual(urllib3_span.p, test_span.s)
        self.assertEqual(pyramid_span.p, urllib3_span.s)

        # Error logging
        self.assertIsNone(test_span.ec)
        self.assertEqual(1, urllib3_span.ec)
        self.assertEqual(1, pyramid_span.ec)

        # HTTP SDK span
        self.assertEqual("sdk", pyramid_span.n)
        self.assertEqual('http', pyramid_span.data["sdk"]["name"])
        self.assertEqual('entry', pyramid_span.data["sdk"]["type"])

        sdk_data = pyramid_span.data["sdk"]["custom"]
        self.assertEqual('127.0.0.1:' + str(testenv['pyramid_port']), sdk_data["tags"]["http.host"])
        self.assertEqual('/exception', sdk_data["tags"]["http.url"])
        self.assertEqual('GET', sdk_data["tags"]["http.method"])
        self.assertEqual(500, sdk_data["tags"]["http.status"])
        self.assertEqual("fake exception", sdk_data["tags"]["message"])
        self.assertNotIn("http.path_tpl", sdk_data["tags"])

        # urllib3
        self.assertEqual("test", test_span.data["sdk"]["name"])
        self.assertEqual("urllib3", urllib3_span.n)
        self.assertEqual(500, urllib3_span.data["http"]["status"])
        self.assertEqual(testenv["pyramid_server"] + '/exception', urllib3_span.data["http"]["url"])
        self.assertEqual("GET", urllib3_span.data["http"]["method"])
        self.assertIsNotNone(urllib3_span.stack)
        self.assertTrue(type(urllib3_span.stack) is list)
        self.assertTrue(len(urllib3_span.stack) > 1)



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2017

from __future__ import absolute_import

from .instrumentation.wsgi import InstanaWSGIMiddleware
from .instrumentation.asgi import InstanaASGIMiddleware


# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

from __future__ import absolute_import

import logging
import unittest
from instana.singletons import tracer


class TestLogging(unittest.TestCase):
    def setUp(self):
        """ Clear all spans before a test run """
        self.recorder = tracer.recorder
        self.recorder.clear_spans()
        self.logger = logging.getLogger('unit test')

    def tearDown(self):
        """ Do nothing for now """
        return None

    def test_no_span(self):
        with tracer.start_active_span('test'):
            self.logger.info('info message')

        spans = self.recorder.queued_spans()
        self.assertEqual(1, len(spans))

    def test_extra_span(self):
        with tracer.start_active_span('test'):
            self.logger.warning('foo %s', 'bar')

        spans = self.recorder.queued_spans()
        self.assertEqual(2, len(spans))
        self.assertEqual(2, spans[0].k)

        self.assertEqual('foo bar', spans[0].data["log"].get('message'))

    def test_log_with_tuple(self):
        with tracer.start_active_span('test'):
            self.logger.warning('foo %s', ("bar",))

        spans = self.recorder.queued_spans()
        self.assertEqual(2, len(spans))
        self.assertEqual(2, spans[0].k)

        self.assertEqual("foo ('bar',)", spans[0].data["log"].get('message'))

    def test_parameters(self):
        with tracer.start_active_span('test'):
            try:
                a = 42
                b = 0
                c = a / b
            except Exception as e:
                self.logger.exception('Exception: %s', str(e))

        spans = self.recorder.queued_spans()
        self.assertEqual(2, len(spans))

        self.assertIsNotNone(spans[0].data["log"].get('parameters'))




# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

"""
Instana ASGI Middleware
"""
import opentracing

from ..log import logger
from ..singletons import async_tracer, agent
from ..util.secrets import strip_secrets_from_query


class InstanaASGIMiddleware:
    """
    Instana ASGI Middleware
    """

    def __init__(self, app):
        self.app = app

    def _extract_custom_headers(self, span, headers):
        try:
            for custom_header in agent.options.extra_http_headers:
                # Headers are in the following format: b'x-header-1'
                for header_pair in headers:
                    if header_pair[0].decode('utf-8').lower() == custom_header.lower():
                        span.set_tag("http.header.%s" % custom_header, header_pair[1].decode('utf-8'))
        except Exception:
            logger.debug("extract_custom_headers: ", exc_info=True)

    def _collect_kvs(self, scope, span):
        try:
            span.set_tag('span.kind', 'entry')
            span.set_tag('http.path', scope.get('path'))
            span.set_tag('http.method', scope.get('method'))

            server = scope.get('server')
            if isinstance(server, tuple):
                span.set_tag('http.host', server[0])

            query = scope.get('query_string')
            if isinstance(query, (str, bytes)) and len(query):
                if isinstance(query, bytes):
                    query = query.decode('utf-8')
                scrubbed_params = strip_secrets_from_query(query, agent.options.secrets_matcher,
                                                           agent.options.secrets_list)
                span.set_tag("http.params", scrubbed_params)

            app = scope.get('app')
            if app is not None and hasattr(app, 'routes'):
                # Attempt to detect the Starlette routes registered.
                # If Starlette isn't present, we harmlessly dump out.
                from starlette.routing import Match
                for route in scope['app'].routes:
                    if route.matches(scope)[0] == Match.FULL:
                        span.set_tag("http.path_tpl", route.path)
        except Exception:
            logger.debug("ASGI collect_kvs: ", exc_info=True)

    async def __call__(self, scope, receive, send):
        request_context = None

        if scope["type"] not in ("http", "websocket"):
            await self.app(scope, receive, send)
            return

        request_headers = scope.get('headers')
        if isinstance(request_headers, list):
            request_context = async_tracer.extract(opentracing.Format.BINARY, request_headers)

        async def send_wrapper(response):
            span = async_tracer.active_span
            if span is None:
                await send(response)
            else:
                if response['type'] == 'http.response.start':
                    try:
                        status_code = response.get('status')
                        if status_code is not None:
                            if 500 <= int(status_code) <= 511:
                                span.mark_as_errored()
                            span.set_tag('http.status_code', status_code)

                        headers = response.get('headers')
                        if headers is not None:
                            async_tracer.inject(span.context, opentracing.Format.BINARY, headers)
                    except Exception:
                        logger.debug("send_wrapper: ", exc_info=True)

                try:
                    await send(response)
                except Exception as exc:
                    span.log_exception(exc)
                    raise

        with async_tracer.start_active_span("asgi", child_of=request_context) as tracing_scope:
            self._collect_kvs(scope, tracing_scope.span)
            if 'headers' in scope and agent.options.extra_http_headers is not None:
                self._extract_custom_headers(tracing_scope.span, scope['headers'])

            try:
                await self.app(scope, receive, send_wrapper)
            except Exception as exc:
                tracing_scope.span.log_exception(exc)
                raise exc



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2021

from instana.propagators.binary_propagator import BinaryPropagator
from instana.span_context import SpanContext
import unittest


class TestBinaryPropagator(unittest.TestCase):
    def setUp(self):
        self.bp = BinaryPropagator()

    def test_inject_carrier_dict(self):
        carrier = {}
        ctx = SpanContext(span_id="1234567890abcdef", trace_id="1234d0e0e4736234",
                                  level=1, baggage={}, sampled=True,
                                  synthetic=False)
        carrier = self.bp.inject(ctx, carrier)
        self.assertEqual(carrier[b'x-instana-t'], b"1234d0e0e4736234")

    def test_inject_carrier_dict_w3c_True(self):
        carrier = {}
        ctx = SpanContext(span_id="1234567890abcdef", trace_id="1234d0e0e4736234",
                                  level=1, baggage={}, sampled=True,
                                  synthetic=False)
        carrier = self.bp.inject(ctx, carrier, disable_w3c_trace_context=False)
        self.assertEqual(carrier[b'x-instana-t'], b"1234d0e0e4736234")
        self.assertEqual(carrier[b'traceparent'], b'00-00000000000000001234d0e0e4736234-1234567890abcdef-01')
        self.assertEqual(carrier[b'tracestate'], b'in=1234d0e0e4736234;1234567890abcdef')

    def test_inject_carrier_list(self):
        carrier = []
        ctx = SpanContext(span_id="1234567890abcdef", trace_id="1234d0e0e4736234",
                                  level=1, baggage={}, sampled=True,
                                  synthetic=False)
        carrier = self.bp.inject(ctx, carrier)
        self.assertEqual(carrier[0], (b'x-instana-t', b'1234d0e0e4736234'))

    def test_inject_carrier_list_w3c_True(self):
        carrier = []
        ctx = SpanContext(span_id="1234567890abcdef", trace_id="1234d0e0e4736234",
                                  level=1, baggage={}, sampled=True,
                                  synthetic=False)
        carrier = self.bp.inject(ctx, carrier, disable_w3c_trace_context=False)
        self.assertEqual(carrier[2], (b'x-instana-t', b'1234d0e0e4736234'))
        self.assertEqual(carrier[0], (b'traceparent', b'00-00000000000000001234d0e0e4736234-1234567890abcdef-01'))
        self.assertEqual(carrier[1], (b'tracestate', b'in=1234d0e0e4736234;1234567890abcdef'))

    def test_inject_carrier_tupple(self):
        carrier = ()
        ctx = SpanContext(span_id="1234567890abcdef", trace_id="1234d0e0e4736234",
                                  level=1, baggage={}, sampled=True,
                                  synthetic=False)
        carrier = self.bp.inject(ctx, carrier)
        self.assertEqual(carrier[0], (b'x-instana-t', b'1234d0e0e4736234'))

    def test_inject_carrier_tupple_w3c_True(self):
        carrier = ()
        ctx = SpanContext(span_id="1234567890abcdef", trace_id="1234d0e0e4736234",
                                  level=1, baggage={}, sampled=True,
                                  synthetic=False)
        carrier = self.bp.inject(ctx, carrier, disable_w3c_trace_context=False)
        self.assertEqual(carrier[2], (b'x-instana-t', b'1234d0e0e4736234'))
        self.assertEqual(carrier[0], (b'traceparent', b'00-00000000000000001234d0e0e4736234-1234567890abcdef-01'))
        self.assertEqual(carrier[1], (b'tracestate', b'in=1234d0e0e4736234;1234567890abcdef'))

    def test_inject_carrier_set_exception(self):
        carrier = set()
        ctx = SpanContext(span_id="1234567890abcdef", trace_id="1234d0e0e4736234",
                                  level=1, baggage={}, sampled=True,
                                  synthetic=False)
        carrier = self.bp.inject(ctx, carrier)
        self.assertIsNone(carrier)